import duckdb
import numpy as np
import os, psutil
import pytest

try:
    import pyarrow as pa

    can_run = True
except ImportError:
    can_run = False
from conftest import NumpyPandas, ArrowPandas


def check_memory(function_to_check, pandas, duckdb_cursor):
    process = psutil.Process(os.getpid())
    mem_usage = process.memory_info().rss / (10**9)
    for __ in range(100):
        function_to_check(pandas, duckdb_cursor)
    cur_mem_usage = process.memory_info().rss / (10**9)
    # This seems a good empirical value
    assert cur_mem_usage / 3 < mem_usage


def from_df(pandas, duckdb_cursor):
    df = pandas.DataFrame({"x": np.random.rand(1_000_000)})
    return duckdb_cursor.from_df(df)


def from_arrow(pandas, duckdb_cursor):
    data = pa.array(np.random.rand(1_000_000), type=pa.float32())
    arrow_table = pa.Table.from_arrays([data], ['a'])
    duckdb_cursor.from_arrow(arrow_table)


def arrow_replacement(pandas, duckdb_cursor):
    data = pa.array(np.random.rand(1_000_000), type=pa.float32())
    arrow_table = pa.Table.from_arrays([data], ['a'])
    duckdb_cursor.query("select sum(a) from arrow_table").fetchall()


def pandas_replacement(pandas, duckdb_cursor):
    df = pandas.DataFrame({"x": np.random.rand(1_000_000)})
    duckdb_cursor.query("select sum(x) from df").fetchall()


class TestRelationDependencyMemoryLeak(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_from_arrow_leak(self, pandas, duckdb_cursor):
        if not can_run:
            return
        check_memory(from_arrow, pandas, duckdb_cursor)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_from_df_leak(self, pandas, duckdb_cursor):
        check_memory(from_df, pandas, duckdb_cursor)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_arrow_replacement_scan_leak(self, pandas, duckdb_cursor):
        if not can_run:
            return
        check_memory(arrow_replacement, pandas, duckdb_cursor)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_pandas_replacement_scan_leak(self, pandas, duckdb_cursor):
        check_memory(pandas_replacement, pandas, duckdb_cursor)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_relation_view_leak(self, pandas, duckdb_cursor):
        rel = from_df(pandas, duckdb_cursor)
        rel.create_view("bla")
        duckdb_cursor.unregister("bla")
        assert rel.query("bla", "select count(*) from bla").fetchone()[0] == 1_000_000



import gc
import duckdb
import pytest
import os, psutil
import pandas as pd


@pytest.fixture
def check_leaks():
    process = psutil.Process(os.getpid())
    gc.collect()
    initial_mem_usage = process.memory_info().rss
    yield
    gc.collect()
    final_mem_usage = process.memory_info().rss
    difference = final_mem_usage - initial_mem_usage
    print("difference:", difference)
    # Assert that the amount of used memory does not pass 5mb
    assert difference <= 5_000_000


class TestMemoryLeaks(object):
    def test_fetchmany(self, duckdb_cursor, check_leaks):
        datetimes = ['1985-01-30T16:41:43' for _ in range(10000)]

        df = pd.DataFrame({'time': pd.Series(data=datetimes)})
        for _ in range(100):
            duckdb_cursor.sql('select time::TIMESTAMP from df').fetchmany(10000)



import duckdb
import pytest
import os
import shutil


class TestWindowsAbsPath(object):
    def test_windows_path_accent(self):
        if os.name != 'nt':
            return
        current_directory = os.getcwd()
        test_dir = os.path.join(current_directory, 'tést')
        if os.path.isdir(test_dir):
            shutil.rmtree(test_dir)
        os.mkdir(test_dir)

        dbname = 'test.db'
        dbpath = os.path.join(test_dir, dbname)
        con = duckdb.connect(dbpath)
        con.execute("CREATE OR REPLACE TABLE int AS SELECT * FROM range(10) t(i)")
        res = con.execute("SELECT COUNT(*) FROM int").fetchall()
        assert res[0][0] == 10
        del res
        del con

        os.chdir('tést')
        dbpath = os.path.join('..', dbpath)
        con = duckdb.connect(dbpath)
        res = con.execute("SELECT COUNT(*) FROM int").fetchall()
        assert res[0][0] == 10
        del res
        del con

        con = duckdb.connect(dbname)
        res = con.execute("SELECT COUNT(*) FROM int").fetchall()
        assert res[0][0] == 10
        del res
        del con

        os.chdir('..')

    def test_windows_abs_path(self):
        if os.name != 'nt':
            return
        current_directory = os.getcwd()
        dbpath = os.path.join(current_directory, 'test.db')
        con = duckdb.connect(dbpath)
        con.execute("CREATE OR REPLACE TABLE int AS SELECT * FROM range(10) t(i)")
        res = con.execute("SELECT COUNT(*) FROM int").fetchall()
        assert res[0][0] == 10
        del res
        del con

        assert dbpath[1] == ':'
        # remove the drive separator and reconnect
        dbpath = dbpath[2:]
        con = duckdb.connect(dbpath)
        res = con.execute("SELECT COUNT(*) FROM int").fetchall()
        assert res[0][0] == 10
        del res
        del con

        # forward slashes work as well
        dbpath = dbpath.replace('\\', '/')
        con = duckdb.connect(dbpath)
        res = con.execute("SELECT COUNT(*) FROM int").fetchall()
        assert res[0][0] == 10
        del res
        del con



import duckdb
import pytest
import datetime

from conftest import pandas_supports_arrow_backend

pd = pytest.importorskip("pandas", '2.0.0')
import numpy as np
from pandas.api.types import is_integer_dtype


@pytest.mark.skipif(not pandas_supports_arrow_backend(), reason="pandas does not support the 'pyarrow' backend")
class TestPandasArrow(object):
    def test_pandas_arrow(self, duckdb_cursor):
        pd = pytest.importorskip("pandas")
        df = pd.DataFrame({'a': pd.Series([5, 4, 3])}).convert_dtypes()
        con = duckdb.connect()
        res = con.sql("select * from df").fetchall()
        assert res == [(5,), (4,), (3,)]

    def test_mixed_columns(self):
        df = pd.DataFrame(
            {
                'strings': pd.Series(['abc', 'DuckDB', 'quack', 'quack']),
                'timestamps': pd.Series(
                    [
                        datetime.datetime(1990, 10, 21),
                        datetime.datetime(2023, 1, 11),
                        datetime.datetime(2001, 2, 5),
                        datetime.datetime(1990, 10, 21),
                    ]
                ),
                'objects': pd.Series([[5, 4, 3], 'test', None, {'a': 42}]),
                'integers': np.ndarray((4,), buffer=np.array([1, 2, 3, 4, 5]), offset=np.int_().itemsize, dtype=int),
            }
        )
        pyarrow_df = df.convert_dtypes(dtype_backend='pyarrow')
        con = duckdb.connect()
        with pytest.raises(
            duckdb.InvalidInputException, match='The dataframe could not be converted to a pyarrow.lib.Table'
        ):
            res = con.sql('select * from pyarrow_df').fetchall()

        numpy_df = pd.DataFrame(
            {'a': np.ndarray((2,), buffer=np.array([1, 2, 3]), offset=np.int_().itemsize, dtype=int)}
        ).convert_dtypes(dtype_backend='numpy_nullable')
        arrow_df = pd.DataFrame(
            {
                'a': pd.Series(
                    [
                        datetime.datetime(1990, 10, 21),
                        datetime.datetime(2023, 1, 11),
                        datetime.datetime(2001, 2, 5),
                        datetime.datetime(1990, 10, 21),
                    ]
                )
            }
        ).convert_dtypes(dtype_backend='pyarrow')
        python_df = pd.DataFrame({'a': pd.Series(['test', [5, 4, 3], {'a': 42}])}).convert_dtypes()

        df = pd.concat([numpy_df['a'], arrow_df['a'], python_df['a']], axis=1, keys=['numpy', 'arrow', 'python'])
        assert is_integer_dtype(df.dtypes['numpy'])
        assert isinstance(df.dtypes['arrow'], pd.ArrowDtype)
        assert isinstance(df.dtypes['python'], np.dtype('O').__class__)

        with pytest.raises(
            duckdb.InvalidInputException, match='The dataframe could not be converted to a pyarrow.lib.Table'
        ):
            res = con.sql('select * from df').fetchall()

    def test_empty_df(self):
        df = pd.DataFrame(
            {
                'string': pd.Series(data=[], dtype='string'),
                'object': pd.Series(data=[], dtype='object'),
                'Int64': pd.Series(data=[], dtype='Int64'),
                'Float64': pd.Series(data=[], dtype='Float64'),
                'bool': pd.Series(data=[], dtype='bool'),
                'datetime64[ns]': pd.Series(data=[], dtype='datetime64[ns]'),
                'datetime64[ms]': pd.Series(data=[], dtype='datetime64[ms]'),
                'datetime64[us]': pd.Series(data=[], dtype='datetime64[us]'),
                'datetime64[s]': pd.Series(data=[], dtype='datetime64[s]'),
                'category': pd.Series(data=[], dtype='category'),
                'timedelta64[ns]': pd.Series(data=[], dtype='timedelta64[ns]'),
            }
        )
        pyarrow_df = df.convert_dtypes(dtype_backend='pyarrow')

        con = duckdb.connect()
        res = con.sql('select * from pyarrow_df').fetchall()
        assert res == []

    def test_completely_null_df(self):
        df = pd.DataFrame(
            {
                'a': pd.Series(
                    data=[
                        None,
                        np.nan,
                        pd.NA,
                    ]
                )
            }
        )
        pyarrow_df = df.convert_dtypes(dtype_backend='pyarrow')

        con = duckdb.connect()
        res = con.sql('select * from pyarrow_df').fetchall()
        assert res == [(None,), (None,), (None,)]

    def test_mixed_nulls(self):
        df = pd.DataFrame(
            {
                'float': pd.Series(data=[4.123123, None, 7.23456], dtype='Float64'),
                'int64': pd.Series(data=[-234234124, 709329413, pd.NA], dtype='Int64'),
                'bool': pd.Series(data=[np.nan, True, False], dtype='boolean'),
                'string': pd.Series(data=['NULL', None, 'quack']),
                'list[str]': pd.Series(data=[['Huey', 'Dewey', 'Louie'], [None, pd.NA, np.nan, 'DuckDB'], None]),
                'datetime64': pd.Series(
                    data=[datetime.datetime(2011, 8, 16, 22, 7, 8), None, datetime.datetime(2010, 4, 26, 18, 14, 14)]
                ),
                'date': pd.Series(data=[datetime.date(2008, 5, 28), datetime.date(2013, 7, 14), None]),
            }
        )
        pyarrow_df = df.convert_dtypes(dtype_backend='pyarrow')
        con = duckdb.connect()
        res = con.sql('select * from pyarrow_df').fetchone()
        assert res == (
            4.123123,
            -234234124,
            None,
            'NULL',
            ['Huey', 'Dewey', 'Louie'],
            datetime.datetime(2011, 8, 16, 22, 7, 8),
            datetime.date(2008, 5, 28),
        )



import duckdb
import pytest
from pytest import mark
import datetime
import decimal
import pytz

pa = pytest.importorskip("pyarrow")

# This causes Arrow to output an array at the very end with an offset
# the parent struct will have an offset that needs to be used when scanning the child array
MAGIC_ARRAY_SIZE = 2**17 + 1


def pa_time32():
    return pa.time32


def pa_time64():
    return pa.time64


def pa_date32():
    return pa.date32


def pa_date64():
    return pa.date64


def pa_timestamp():
    return pa.timestamp


def pa_duration():
    return pa.duration


def pa_month_day_nano_interval():
    return pa.month_day_nano_interval


def month_interval(months):
    return (months, 0, 0)


def day_interval(days):
    return (0, days, 0)


def nano_interval(nanos):
    return (0, 0, nanos)


def increment_or_null(val: str, increment):
    if not val:
        return val
    return str(int(val) + increment)


def decimal_value(value, precision, scale):
    val = str(value)
    actual_width = precision - scale
    if len(val) > actual_width:
        return decimal.Decimal('9' * actual_width)
    return decimal.Decimal(val)


def expected_result(col1_null, col2_null, expected):
    col1 = None if col1_null else expected
    if col1_null or col2_null:
        col2 = None
    else:
        col2 = expected
    return [(col1, col2)]


test_nulls = lambda: mark.parametrize(
    ['col1_null', 'col2_null'], [(False, True), (True, False), (True, True), (False, False)]
)


class TestArrowOffsets(object):
    @test_nulls()
    def test_struct_of_strings(self, duckdb_cursor, col1_null, col2_null):
        col1 = [str(i) for i in range(0, MAGIC_ARRAY_SIZE)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": i} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", pa.string()), ("col2", pa.struct({"a": pa.string()}))]),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, '131072')

    @test_nulls()
    def test_struct_of_bools(self, duckdb_cursor, col1_null, col2_null):
        tuples = [False for i in range(0, MAGIC_ARRAY_SIZE)]
        tuples[-1] = True

        col1 = tuples
        if col1_null:
            col1[-1] = None
        col2 = [{"a": i} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", pa.bool_()), ("col2", pa.struct({"a": pa.bool_()}))]),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, True)

    @pytest.mark.parametrize(
        ["constructor", "expected"],
        [
            (pa_date32(), datetime.date(2328, 11, 12)),
            (pa_date64(), datetime.date(1970, 1, 1)),
        ],
    )
    @test_nulls()
    def test_struct_of_dates(self, duckdb_cursor, constructor, expected, col1_null, col2_null):
        tuples = [i for i in range(0, MAGIC_ARRAY_SIZE)]

        col1 = tuples
        if col1_null:
            col1[-1] = None
        col2 = [{"a": i} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", constructor()), ("col2", pa.struct({"a": constructor()}))]),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, expected)

    @test_nulls()
    def test_struct_of_enum(self, duckdb_cursor, col1_null, col2_null):
        enum_type = pa.dictionary(pa.int64(), pa.utf8())

        tuples = ['red' for i in range(MAGIC_ARRAY_SIZE)]
        tuples[-1] = 'green'
        if col1_null:
            tuples[-1] = None

        struct_tuples = [{"a": x} for x in tuples]
        if col2_null:
            struct_tuples[-1] = None

        arrow_table = pa.Table.from_pydict(
            {'col1': pa.array(tuples, enum_type), 'col2': pa.array(struct_tuples, pa.struct({"a": enum_type}))},
            schema=pa.schema([("col1", enum_type), ("col2", pa.struct({"a": enum_type}))]),
        )
        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, 'green')

    @test_nulls()
    def test_struct_of_blobs(self, duckdb_cursor, col1_null, col2_null):
        col1 = [str(i) for i in range(0, MAGIC_ARRAY_SIZE)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": i} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", pa.binary()), ("col2", pa.struct({"a": pa.binary()}))]),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, b'131072')

    @test_nulls()
    @pytest.mark.parametrize(
        ["constructor", "unit", "expected"],
        [
            (pa_time32(), 'ms', datetime.time(0, 2, 11, 72000)),
            (pa_time32(), 's', datetime.time(23, 59, 59)),
            (pa_time64(), 'ns', datetime.time(0, 0, 0, 131)),
            (pa_time64(), 'us', datetime.time(0, 0, 0, 131072)),
        ],
    )
    def test_struct_of_time(self, duckdb_cursor, constructor, unit, expected, col1_null, col2_null):
        size = MAGIC_ARRAY_SIZE
        if unit == 's':
            # FIXME: We limit the size because we don't support time values > 24 hours
            size = 86400  # The amount of seconds in a day

        col1 = [i for i in range(0, size)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": i} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", constructor(unit)), ("col2", pa.struct({"a": constructor(unit)}))]),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {size-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, expected)

    @test_nulls()
    # NOTE: there is sadly no way to create a 'interval[months]' (tiM) type from pyarrow
    @pytest.mark.parametrize(
        ["constructor", "expected", "converter"],
        [
            (pa_month_day_nano_interval(), datetime.timedelta(days=3932160), month_interval),
            (pa_month_day_nano_interval(), datetime.timedelta(days=131072), day_interval),
            (pa_month_day_nano_interval(), datetime.timedelta(microseconds=131), nano_interval),
        ],
    )
    def test_struct_of_interval(self, duckdb_cursor, constructor, expected, converter, col1_null, col2_null):
        size = MAGIC_ARRAY_SIZE

        col1 = [converter(i) for i in range(0, size)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": i} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", constructor()), ("col2", pa.struct({"a": constructor()}))]),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {size-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, expected)

    @test_nulls()
    @pytest.mark.parametrize(
        ["constructor", "unit", "expected"],
        [
            (pa_duration(), 'ms', datetime.timedelta(seconds=131, microseconds=72000)),
            (pa_duration(), 's', datetime.timedelta(days=1, seconds=44672)),
            (pa_duration(), 'ns', datetime.timedelta(microseconds=131)),
            (pa_duration(), 'us', datetime.timedelta(microseconds=131072)),
        ],
    )
    def test_struct_of_duration(self, duckdb_cursor, constructor, unit, expected, col1_null, col2_null):
        size = MAGIC_ARRAY_SIZE

        col1 = [i for i in range(0, size)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": i} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", constructor(unit)), ("col2", pa.struct({"a": constructor(unit)}))]),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {size-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, expected)

    @test_nulls()
    @pytest.mark.parametrize(
        ["constructor", "unit", "expected"],
        [
            (pa_timestamp(), 'ms', datetime.datetime(1970, 1, 1, 0, 2, 11, 72000, tzinfo=pytz.utc)),
            (pa_timestamp(), 's', datetime.datetime(1970, 1, 2, 12, 24, 32, 0, tzinfo=pytz.utc)),
            (pa_timestamp(), 'ns', datetime.datetime(1970, 1, 1, 0, 0, 0, 131, tzinfo=pytz.utc)),
            (pa_timestamp(), 'us', datetime.datetime(1970, 1, 1, 0, 0, 0, 131072, tzinfo=pytz.utc)),
        ],
    )
    def test_struct_of_timestamp_tz(self, duckdb_cursor, constructor, unit, expected, col1_null, col2_null):
        size = MAGIC_ARRAY_SIZE

        duckdb_cursor.execute("set timezone='UTC'")
        col1 = [i for i in range(0, size)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": i} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema(
                [("col1", constructor(unit, 'UTC')), ("col2", pa.struct({"a": constructor(unit, 'UTC')}))]
            ),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {size-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, expected)

    @test_nulls()
    def test_struct_of_large_blobs(self, duckdb_cursor, col1_null, col2_null):
        col1 = [str(i) for i in range(0, MAGIC_ARRAY_SIZE)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": i} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", pa.large_binary()), ("col2", pa.struct({"a": pa.large_binary()}))]),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, b'131072')

    @test_nulls()
    @pytest.mark.parametrize(
        ["precision_scale", "expected"],
        [
            ((38, 37), decimal.Decimal('9.0000000000000000000000000000000000000')),
            ((38, 24), decimal.Decimal('131072.000000000000000000000000')),
            ((18, 14), decimal.Decimal('9999.00000000000000')),
            ((18, 5), decimal.Decimal('131072.00000')),
            ((9, 7), decimal.Decimal('99.0000000')),
            ((9, 3), decimal.Decimal('131072.000')),
            ((4, 2), decimal.Decimal('99.00')),
            ((4, 0), decimal.Decimal('9999')),
        ],
    )
    def test_struct_of_decimal(self, duckdb_cursor, precision_scale, expected, col1_null, col2_null):
        precision, scale = precision_scale
        col1 = [decimal_value(i, precision, scale) for i in range(0, MAGIC_ARRAY_SIZE)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": i} for i in col1]
        if col2_null:
            col2[-1] = None

        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema(
                [("col1", pa.decimal128(precision, scale)), ("col2", pa.struct({"a": pa.decimal128(precision, scale)}))]
            ),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        assert res == expected_result(col1_null, col2_null, expected)

    @test_nulls()
    def test_struct_of_small_list(self, duckdb_cursor, col1_null, col2_null):
        col1 = [str(i) for i in range(0, MAGIC_ARRAY_SIZE)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": [i, i, i]} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", pa.string()), ("col2", pa.struct({"a": pa.list_(pa.string())}))]),
        )
        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        res1 = None if col1_null else '131072'
        if col2_null:
            res2 = None
        elif col1_null:
            res2 = [None, None, None]
        else:
            res2 = ['131072', '131072', '131072']
        assert res == [(res1, res2)]

    @test_nulls()
    def test_struct_of_fixed_size_list(self, duckdb_cursor, col1_null, col2_null):
        col1 = [str(i) for i in range(0, MAGIC_ARRAY_SIZE)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": [i, i, i]} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", pa.string()), ("col2", pa.struct({"a": pa.list_(pa.string(), 3)}))]),
        )
        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        res1 = None if col1_null else '131072'
        if col2_null:
            res2 = None
        elif col1_null:
            res2 = (None, None, None)
        else:
            res2 = ('131072', '131072', '131072')
        assert res == [(res1, res2)]

    @test_nulls()
    def test_struct_of_fixed_size_blob(self, duckdb_cursor, col1_null, col2_null):
        col1 = [str(i) for i in range(0, MAGIC_ARRAY_SIZE)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": [i, increment_or_null(i, 1), increment_or_null(i, 2)]} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", pa.binary()), ("col2", pa.struct({"a": pa.list_(pa.binary(), 3)}))]),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        res1 = None if col1_null else b'131072'
        if col2_null:
            res2 = None
        elif col1_null:
            res2 = (None, None, None)
        else:
            res2 = (b'131072', b'131073', b'131074')
        assert res == [(res1, res2)]

    @test_nulls()
    def test_struct_of_list_of_blobs(self, duckdb_cursor, col1_null, col2_null):
        col1 = [str(i) for i in range(0, MAGIC_ARRAY_SIZE)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": [i, increment_or_null(i, 1), increment_or_null(i, 2)]} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            schema=pa.schema([("col1", pa.binary()), ("col2", pa.struct({"a": pa.list_(pa.binary())}))]),
        )

        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        res1 = None if col1_null else b'131072'
        if col2_null:
            res2 = None
        elif col1_null:
            res2 = [None, None, None]
        else:
            res2 = [b'131072', b'131073', b'131074']
        assert res == [(res1, res2)]

    @test_nulls()
    def test_struct_of_list_of_list(self, duckdb_cursor, col1_null, col2_null):
        col1 = [i for i in range(0, MAGIC_ARRAY_SIZE)]
        if col1_null:
            col1[-1] = None
        # "a" in the struct matches the value for col1
        col2 = [{"a": [[i, i, i], [], None, [i]]} for i in col1]
        if col2_null:
            col2[-1] = None
        arrow_table = pa.Table.from_pydict(
            {"col1": col1, "col2": col2},
            # thanks formatter
            schema=pa.schema([("col1", pa.int32()), ("col2", pa.struct({"a": pa.list_(pa.list_(pa.int32()))}))]),
        )
        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1,
                col2.a
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        res1 = None if col1_null else 131072
        if col2_null:
            res2 = None
        elif col1_null:
            res2 = [[None, None, None], [], None, [None]]
        else:
            res2 = [[131072, 131072, 131072], [], None, [131072]]
        assert res == [(res1, res2)]

    @pytest.mark.parametrize('col1_null', [True, False])
    def test_list_of_struct(self, duckdb_cursor, col1_null):
        # One single tuple containing a very big list
        tuples = [{"a": i} for i in range(0, MAGIC_ARRAY_SIZE)]
        if col1_null:
            tuples[-1] = None
        tuples = [tuples]
        arrow_table = pa.Table.from_pydict(
            {"col1": tuples},
            schema=pa.schema([("col1", pa.list_(pa.struct({"a": pa.int32()})))]),
        )
        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1
            FROM arrow_table
        """
        ).fetchall()
        res = res[0][0]
        for i, x in enumerate(res[:-1]):
            assert x.__class__ == dict
            assert x['a'] == i
        if col1_null:
            assert res[-1] == None
        else:
            assert res[-1]['a'] == len(res) - 1

    @pytest.mark.parametrize(['outer_null', 'inner_null'], [(True, False), (False, True)])
    def test_list_of_list_of_struct(self, duckdb_cursor, outer_null, inner_null):
        tuples = [[[{"a": str(i), "b": None, "c": [i]}]] for i in range(MAGIC_ARRAY_SIZE)]
        if outer_null:
            tuples[-1] = None
        else:
            inner = [[{"a": 'aaaaaaaaaaaaaaa', "b": 'test', "c": [1, 2, 3]}] for _ in range(MAGIC_ARRAY_SIZE)]
            if inner_null:
                inner[-1] = None
            tuples[-1] = inner

        # MAGIC_ARRAY_SIZE tuples, all containing a single child list
        # except the last tuple, the child list of which is also MAGIC_ARRAY_SIZE in length
        arrow_table = pa.Table.from_pydict(
            {"col1": tuples},
            schema=pa.schema(
                [
                    (
                        "col1",
                        pa.list_(
                            pa.list_(pa.struct([("a", pa.string()), ("b", pa.string()), ("c", pa.list_(pa.int32()))]))
                        ),
                    )
                ]
            ),
        )
        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1
            FROM arrow_table OFFSET {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchall()
        if outer_null:
            assert res == [(None,)]
        else:
            if inner_null:
                assert res[-1][-1][-1] == None
            else:
                assert res[-1][-1][-1] == 131072

    @pytest.mark.parametrize('col1_null', [True, False])
    def test_struct_of_list(self, duckdb_cursor, col1_null):
        # All elements are of size 1
        tuples = [{"a": [str(i)]} for i in range(MAGIC_ARRAY_SIZE)]
        if col1_null:
            tuples[-1] = None
        else:
            tuples[-1] = {"a": [str(x) for x in range(MAGIC_ARRAY_SIZE)]}

        # Except the very last element, which is big

        arrow_table = pa.Table.from_pydict(
            {"col1": tuples}, schema=pa.schema([("col1", pa.struct({"a": pa.list_(pa.string())}))])
        )
        res = duckdb_cursor.sql(
            f"""
            SELECT
                col1
            FROM arrow_table offset {MAGIC_ARRAY_SIZE-1}
        """
        ).fetchone()
        if col1_null:
            assert res[0] == None
        else:
            assert res[0]['a'][-1] == '131072'



import duckdb
import pandas as pd
import numpy
import datetime
import time


class TestProgressBarPandas(object):
    def test_progress_pandas_single(self, duckdb_cursor):
        con = duckdb.connect()
        df = pd.DataFrame({'i': numpy.arange(10000000)})

        con.register('df', df)
        con.register('df_2', df)
        con.execute("PRAGMA progress_bar_time=1")
        con.execute("PRAGMA disable_print_progress_bar")
        result = con.execute("SELECT SUM(df.i) FROM df inner join df_2 on (df.i = df_2.i)").fetchall()
        assert result[0][0] == 49999995000000

    def test_progress_pandas_parallel(self, duckdb_cursor):
        con = duckdb.connect()
        df = pd.DataFrame({'i': numpy.arange(10000000)})

        con.register('df', df)
        con.register('df_2', df)
        con.execute("PRAGMA progress_bar_time=1")
        con.execute("PRAGMA disable_print_progress_bar")
        con.execute("PRAGMA threads=4")
        parallel_results = con.execute("SELECT SUM(df.i) FROM df inner join df_2 on (df.i = df_2.i)").fetchall()
        assert parallel_results[0][0] == 49999995000000

    def test_progress_pandas_empty(self, duckdb_cursor):
        con = duckdb.connect()
        df = pd.DataFrame({'i': []})
        con.register('df', df)
        con.execute("PRAGMA progress_bar_time=1")
        con.execute("PRAGMA disable_print_progress_bar")
        result = con.execute("SELECT SUM(df.i) from df").fetchall()
        assert result[0][0] == None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import duckdb
import numpy
import pytest
from conftest import NumpyPandas, ArrowPandas


# Join from pandas not matching identical strings #1767
class TestIssue1767(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_unicode_join_pandas(self, duckdb_cursor, pandas):
        A = pandas.DataFrame({"key": ["a", "п"]})
        B = pandas.DataFrame({"key": ["a", "п"]})
        con = duckdb.connect(":memory:")
        arrow = con.register("A", A).register("B", B)
        q = arrow.query("""SELECT key FROM "A" FULL JOIN "B" USING ("key") ORDER BY key""")
        result = q.df()

        d = {'key': ["a", "п"]}
        df = pandas.DataFrame(data=d)
        pandas.testing.assert_frame_equal(result, df)



from .. import DuckDBPyConnection
from typing import List

SQLNULL: DuckDBPyType
BOOLEAN: DuckDBPyType
TINYINT: DuckDBPyType
UTINYINT: DuckDBPyType
SMALLINT: DuckDBPyType
USMALLINT: DuckDBPyType
INTEGER: DuckDBPyType
UINTEGER: DuckDBPyType
BIGINT: DuckDBPyType
UBIGINT: DuckDBPyType
HUGEINT: DuckDBPyType
UHUGEINT: DuckDBPyType
UUID: DuckDBPyType
FLOAT: DuckDBPyType
DOUBLE: DuckDBPyType
DATE: DuckDBPyType
TIMESTAMP: DuckDBPyType
TIMESTAMP_MS: DuckDBPyType
TIMESTAMP_NS: DuckDBPyType
TIMESTAMP_S: DuckDBPyType
TIME: DuckDBPyType
TIME_TZ: DuckDBPyType
TIMESTAMP_TZ: DuckDBPyType
VARCHAR: DuckDBPyType
BLOB: DuckDBPyType
BIT: DuckDBPyType
INTERVAL: DuckDBPyType

class DuckDBPyType:
    def __init__(self, type_str: str, connection: DuckDBPyConnection = ...) -> None: ...
    def __repr__(self) -> str: ...
    def __eq__(self, other) -> bool: ...
    def __getattr__(self, name: str): DuckDBPyType
    def __getitem__(self, name: str): DuckDBPyType


import duckdb


class Test6315(object):
    def test_6315(self, duckdb_cursor):
        # segfault when accessing description after fetching rows
        c = duckdb.connect(":memory:")
        rv = c.execute("select * from sqlite_master where type = 'table'")
        rv.fetchall()
        desc = rv.description
        names = [x[0] for x in desc]
        assert names == ['type', 'name', 'tbl_name', 'rootpage', 'sql']

        # description of relation
        rel = c.sql("select * from sqlite_master where type = 'table'")
        desc = rel.description
        names = [x[0] for x in desc]
        assert names == ['type', 'name', 'tbl_name', 'rootpage', 'sql']

        rel.fetchall()
        desc = rel.description
        names = [x[0] for x in desc]
        assert names == ['type', 'name', 'tbl_name', 'rootpage', 'sql']



import logging
import sys
from pathlib import Path
from shutil import copyfileobj
from typing import Callable, List
from os.path import exists
from pathlib import PurePosixPath

import duckdb
from duckdb import DuckDBPyConnection, InvalidInputException
from pytest import raises, importorskip, fixture, MonkeyPatch, mark

importorskip('fsspec', '2022.11.0')
from fsspec import filesystem, AbstractFileSystem
from fsspec.implementations.memory import MemoryFileSystem
from fsspec.implementations.local import LocalFileOpener

FILENAME = 'integers.csv'

logging.basicConfig(level=logging.DEBUG)


def intercept(monkeypatch: MonkeyPatch, obj: object, name: str) -> List[str]:
    error_occured = []
    orig = getattr(obj, name)

    def ceptor(*args, **kwargs):
        try:
            return orig(*args, **kwargs)
        except Exception as e:
            error_occured.append(e)
            raise e

    monkeypatch.setattr(obj, name, ceptor)
    return error_occured


@fixture()
def duckdb_cursor():
    with duckdb.connect() as conn:
        yield conn


@fixture()
def memory():
    fs = filesystem('memory', skip_instance_cache=True)

    # ensure each instance is independent (to work around a weird quirk in fsspec)
    fs.store = {}
    fs.pseudo_dirs = ['']

    # copy csv into memory filesystem
    add_file(fs)
    return fs


def add_file(fs, filename=FILENAME):
    with (Path(__file__).parent / 'data' / filename).open('rb') as source, fs.open(filename, 'wb') as dest:
        copyfileobj(source, dest)


class TestPythonFilesystem:
    def test_unregister_non_existent_filesystem(self, duckdb_cursor: DuckDBPyConnection):
        with raises(InvalidInputException):
            duckdb_cursor.unregister_filesystem('fake')

    def test_memory_filesystem(self, duckdb_cursor: DuckDBPyConnection, memory: AbstractFileSystem):
        duckdb_cursor.register_filesystem(memory)

        assert memory.protocol == 'memory'

        duckdb_cursor.execute(f"select * from 'memory://{FILENAME}'")

        assert duckdb_cursor.fetchall() == [(1, 10, 0), (2, 50, 30)]

        duckdb_cursor.unregister_filesystem('memory')

    def test_reject_abstract_filesystem(self, duckdb_cursor: DuckDBPyConnection):
        with raises(InvalidInputException):
            duckdb_cursor.register_filesystem(AbstractFileSystem())

    def test_unregister_builtin(self, require: Callable[[str], DuckDBPyConnection]):
        duckdb_cursor = require('httpfs')
        assert duckdb_cursor.filesystem_is_registered('S3FileSystem') == True
        duckdb_cursor.unregister_filesystem('S3FileSystem')
        assert duckdb_cursor.filesystem_is_registered('S3FileSystem') == False

    def test_multiple_protocol_filesystems(self, duckdb_cursor: DuckDBPyConnection):
        class ExtendedMemoryFileSystem(MemoryFileSystem):
            protocol = ('file', 'local')
            # defer to the original implementation that doesn't hardcode the protocol
            _strip_protocol = classmethod(AbstractFileSystem._strip_protocol.__func__)

        memory = ExtendedMemoryFileSystem(skip_instance_cache=True)
        add_file(memory)
        duckdb_cursor.register_filesystem(memory)
        for protocol in memory.protocol:
            duckdb_cursor.execute(f"select * from '{protocol}://{FILENAME}'")

            assert duckdb_cursor.fetchall() == [(1, 10, 0), (2, 50, 30)]

    def test_write(self, duckdb_cursor: DuckDBPyConnection, memory: AbstractFileSystem):
        duckdb_cursor.register_filesystem(memory)

        duckdb_cursor.execute("copy (select 1) to 'memory://01.csv' (FORMAT CSV, HEADER 0)")

        assert memory.open('01.csv').read() == b'1\n'

    def test_null_bytes(self, duckdb_cursor: DuckDBPyConnection, memory: AbstractFileSystem):
        with memory.open('test.csv', 'wb') as fh:
            fh.write(b'hello\n\0world\0')
        duckdb_cursor.register_filesystem(memory)

        duckdb_cursor.execute('select * from read_csv("memory://test.csv", header = 0)')

        assert duckdb_cursor.fetchall() == [('hello',), ('\0world\0',)]

    def test_read_parquet(self, duckdb_cursor: DuckDBPyConnection, memory: AbstractFileSystem):
        filename = 'binary_string.parquet'
        add_file(memory, filename)

        duckdb_cursor.register_filesystem(memory)

        duckdb_cursor.execute(f"select * from read_parquet('memory://{filename}')")

        assert duckdb_cursor.fetchall() == [(b'foo',), (b'bar',), (b'baz',)]

    def test_write_parquet(self, duckdb_cursor: DuckDBPyConnection, memory: AbstractFileSystem):
        duckdb_cursor.register_filesystem(memory)
        filename = 'output.parquet'

        duckdb_cursor.execute(f'''COPY (SELECT 1) TO 'memory://{filename}' (FORMAT PARQUET);''')

        assert memory.open(filename).read().startswith(b'PAR1')

    def test_when_fsspec_not_installed(self, duckdb_cursor: DuckDBPyConnection, monkeypatch: MonkeyPatch):
        monkeypatch.setitem(sys.modules, 'fsspec', None)

        with raises(ModuleNotFoundError):
            duckdb_cursor.register_filesystem(None)

    @mark.skipif(sys.version_info < (3, 8), reason="ArrowFSWrapper requires python 3.8 or higher")
    def test_arrow_fs_wrapper(self, tmp_path: Path, duckdb_cursor: DuckDBPyConnection):
        fs = importorskip('pyarrow.fs')
        from fsspec.implementations.arrow import ArrowFSWrapper

        local = fs.LocalFileSystem()
        local_fsspec = ArrowFSWrapper(local, skip_instance_cache=True)
        # posix calls here required as ArrowFSWrapper only supports url-like paths (not Windows paths)
        filename = str(PurePosixPath(tmp_path.as_posix()) / "test.csv")
        with local_fsspec.open(filename, mode='w') as f:
            f.write("a,b,c\n")
            f.write("1,2,3\n")
            f.write("4,5,6\n")

        duckdb_cursor.register_filesystem(local_fsspec)
        duckdb_cursor.execute(f"select * from read_csv_auto('local://{filename}', header=true)")

        assert duckdb_cursor.fetchall() == [(1, 2, 3), (4, 5, 6)]

    def test_database_attach(self, tmp_path: Path, monkeypatch: MonkeyPatch):
        db_path = str(tmp_path / 'hello.db')

        # setup a database to attach later
        with duckdb.connect(db_path) as conn:
            conn.execute(
                '''
                CREATE TABLE t (id int);
                INSERT INTO t VALUES (0)
            '''
            )

        assert exists(db_path)

        with duckdb.connect() as conn:
            fs = filesystem('file', skip_instance_cache=True)
            write_errors = intercept(monkeypatch, LocalFileOpener, 'write')
            conn.register_filesystem(fs)
            db_path_posix = str(PurePosixPath(tmp_path.as_posix()) / "hello.db")
            conn.execute(f"ATTACH 'file://{db_path_posix}'")

            conn.execute('INSERT INTO hello.t VALUES (1)')

            conn.execute('FROM hello.t')
            assert conn.fetchall() == [(0,), (1,)]

        # duckdb sometimes seems to swallow write errors, so we use this to ensure that
        # isn't happening
        assert not write_errors

    def test_copy_partition(self, duckdb_cursor: DuckDBPyConnection, memory: AbstractFileSystem):
        duckdb_cursor.register_filesystem(memory)

        duckdb_cursor.execute("copy (select 1 as a) to 'memory://root' (partition_by (a), HEADER 0)")

        assert memory.open('/root/a=1/data_0.csv').read() == b'1\n'

    def test_read_hive_partition(self, duckdb_cursor: DuckDBPyConnection, memory: AbstractFileSystem):
        duckdb_cursor.register_filesystem(memory)
        duckdb_cursor.execute("copy (select 2 as a) to 'memory://partition' (partition_by (a), HEADER 0)")

        path = 'memory:///partition/*/*.csv'

        query = "SELECT * FROM read_csv_auto('" + path + "'"

        # hive partitioning
        duckdb_cursor.execute(query + ', HIVE_PARTITIONING=1' + ');')
        assert duckdb_cursor.fetchall() == [(2, 2)]

        # hive partitioning: auto detection
        duckdb_cursor.execute(query + ');')
        assert duckdb_cursor.fetchall() == [(2, 2)]

        # hive partitioning: cast to int
        duckdb_cursor.execute(query + ', HIVE_PARTITIONING=1' + ', HIVE_TYPES_AUTOCAST=1' + ');')
        assert duckdb_cursor.fetchall() == [(2, 2)]

        # hive partitioning: no cast to int
        duckdb_cursor.execute(query + ', HIVE_PARTITIONING=1' + ', HIVE_TYPES_AUTOCAST=0' + ');')
        assert duckdb_cursor.fetchall() == [(2, '2')]



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import (
    LongType,
    StructType,
    BooleanType,
    StructField,
    StringType,
    IntegerType,
    LongType,
    Row,
    ArrayType,
    MapType,
)
from duckdb.experimental.spark.sql.functions import col, struct, when, lit, array_contains
import duckdb
import re


class TestDataFrameOrderBy(object):
    def test_order_by(self, spark):
        simpleData = [
            ("James", "Sales", "NY", 90000, 34, 10000),
            ("Michael", "Sales", "NY", 86000, 56, 20000),
            ("Robert", "Sales", "CA", 81000, 30, 23000),
            ("Maria", "Finance", "CA", 90000, 24, 23000),
            ("Raman", "Finance", "CA", 99000, 40, 24000),
            ("Scott", "Finance", "NY", 83000, 36, 19000),
            ("Jen", "Finance", "NY", 79000, 53, 15000),
            ("Jeff", "Marketing", "CA", 80000, 25, 18000),
            ("Kumar", "Marketing", "NY", 91000, 50, 21000),
        ]
        columns = ["employee_name", "department", "state", "salary", "age", "bonus"]
        df = spark.createDataFrame(data=simpleData, schema=columns)

        df2 = df.sort("department", "state")
        res1 = df2.collect()
        assert res1 == [
            Row(employee_name='Maria', department='Finance', state='CA', salary=90000, age=24, bonus=23000),
            Row(employee_name='Raman', department='Finance', state='CA', salary=99000, age=40, bonus=24000),
            Row(employee_name='Scott', department='Finance', state='NY', salary=83000, age=36, bonus=19000),
            Row(employee_name='Jen', department='Finance', state='NY', salary=79000, age=53, bonus=15000),
            Row(employee_name='Jeff', department='Marketing', state='CA', salary=80000, age=25, bonus=18000),
            Row(employee_name='Kumar', department='Marketing', state='NY', salary=91000, age=50, bonus=21000),
            Row(employee_name='Robert', department='Sales', state='CA', salary=81000, age=30, bonus=23000),
            Row(employee_name='James', department='Sales', state='NY', salary=90000, age=34, bonus=10000),
            Row(employee_name='Michael', department='Sales', state='NY', salary=86000, age=56, bonus=20000),
        ]

        df2 = df.sort(col("department"), col("state"))
        res2 = df2.collect()
        assert res2 == res1

        df2 = df.orderBy(col("department").asc(), col("state").asc())
        res3 = df2.collect()
        assert res3 == res1

        df2 = df.sort(df.department.asc(), df.state.desc())
        res1 = df2.collect()
        assert res1 == [
            Row(employee_name='Scott', department='Finance', state='NY', salary=83000, age=36, bonus=19000),
            Row(employee_name='Jen', department='Finance', state='NY', salary=79000, age=53, bonus=15000),
            Row(employee_name='Maria', department='Finance', state='CA', salary=90000, age=24, bonus=23000),
            Row(employee_name='Raman', department='Finance', state='CA', salary=99000, age=40, bonus=24000),
            Row(employee_name='Kumar', department='Marketing', state='NY', salary=91000, age=50, bonus=21000),
            Row(employee_name='Jeff', department='Marketing', state='CA', salary=80000, age=25, bonus=18000),
            Row(employee_name='James', department='Sales', state='NY', salary=90000, age=34, bonus=10000),
            Row(employee_name='Michael', department='Sales', state='NY', salary=86000, age=56, bonus=20000),
            Row(employee_name='Robert', department='Sales', state='CA', salary=81000, age=30, bonus=23000),
        ]

        df2 = df.sort(col("department").asc(), col("state").desc())
        res2 = df2.collect()
        assert res2 == res1

        df2 = df.orderBy(col("department").asc(), col("state").desc())
        res3 = df2.collect()
        assert res3 == res1

        df.createOrReplaceTempView("EMP")
        df2 = spark.sql(
            """
            select
                employee_name,
                department,
                state,
                salary,
                age,
                bonus
            from EMP ORDER BY department asc
        """
        )
        res = df2.collect()
        assert res == [
            Row(employee_name='Maria', department='Finance', state='CA', salary=90000, age=24, bonus=23000),
            Row(employee_name='Raman', department='Finance', state='CA', salary=99000, age=40, bonus=24000),
            Row(employee_name='Scott', department='Finance', state='NY', salary=83000, age=36, bonus=19000),
            Row(employee_name='Jen', department='Finance', state='NY', salary=79000, age=53, bonus=15000),
            Row(employee_name='Jeff', department='Marketing', state='CA', salary=80000, age=25, bonus=18000),
            Row(employee_name='Kumar', department='Marketing', state='NY', salary=91000, age=50, bonus=21000),
            Row(employee_name='James', department='Sales', state='NY', salary=90000, age=34, bonus=10000),
            Row(employee_name='Michael', department='Sales', state='NY', salary=86000, age=56, bonus=20000),
            Row(employee_name='Robert', department='Sales', state='CA', salary=81000, age=30, bonus=23000),
        ]



import duckdb
import os
import pytest

try:
    import pyarrow as pa
    import pyarrow.dataset as ds

    can_run = True
except:
    can_run = False


class TestArrowProjectionPushdown(object):
    def test_projection_pushdown_no_filter(self, duckdb_cursor):
        if not can_run:
            return
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("CREATE TABLE test (a  INTEGER, b INTEGER, c INTEGER)")
        duckdb_conn.execute("INSERT INTO  test VALUES (1,1,1),(10,10,10),(100,10,100),(NULL,NULL,NULL)")
        duck_tbl = duckdb_conn.table("test")
        arrow_table = duck_tbl.arrow()
        duckdb_conn.register("testarrowtable", arrow_table)
        assert duckdb_conn.execute("SELECT sum(a) FROM  testarrowtable").fetchall() == [(111,)]

        arrow_dataset = ds.dataset(arrow_table)
        duckdb_conn.register("testarrowdataset", arrow_dataset)
        assert duckdb_conn.execute("SELECT sum(a) FROM  testarrowdataset").fetchall() == [(111,)]



from typing import TYPE_CHECKING, Optional, Union
from .types import StructType

if TYPE_CHECKING:
    from .dataframe import DataFrame
    from .session import SparkSession

PrimitiveType = Union[bool, float, int, str]
OptionalPrimitiveType = Optional[PrimitiveType]


class DataStreamWriter:
    def __init__(self, dataframe: "DataFrame"):
        self.dataframe = dataframe

    def toTable(self, table_name: str) -> None:
        # Should we register the dataframe or create a table from the contents?
        raise NotImplementedError


class DataStreamReader:
    def __init__(self, session: "SparkSession"):
        self.session = session

    def load(
        self,
        path: Optional[str] = None,
        format: Optional[str] = None,
        schema: Union[StructType, str, None] = None,
        **options: OptionalPrimitiveType
    ) -> "DataFrame":
        from duckdb.experimental.spark.sql.dataframe import DataFrame

        raise NotImplementedError


__all__ = ["DataStreamReader", "DataStreamWriter"]



import duckdb
import os

try:
    import pyarrow as pa

    can_run = True
except:
    can_run = False


class Test2426(object):
    def test_2426(self, duckdb_cursor):
        if not can_run:
            return

        con = duckdb.connect()
        con.execute("Create Table test (a integer)")

        for i in range(1024):
            for j in range(2):
                con.execute("Insert Into test values ('" + str(i) + "')")
        con.execute("Insert Into test values ('5000')")
        con.execute("Insert Into test values ('6000')")
        sql = '''
        SELECT  a, COUNT(*) AS repetitions
        FROM    test
        GROUP BY a
        '''

        result_df = con.execute(sql).df()

        arrow_table = con.execute(sql).fetch_arrow_table()

        arrow_df = arrow_table.to_pandas()
        assert result_df['repetitions'].sum() == arrow_df['repetitions'].sum()



# multiple result sets

import numpy
import duckdb


class TestMultipleResultSets(object):
    def test_regular_selection(self, duckdb_cursor, integers):
        duckdb_cursor.execute('SELECT * FROM integers')
        duckdb_cursor.execute('SELECT * FROM integers')
        result = duckdb_cursor.fetchall()
        assert result == [
            (0,),
            (1,),
            (2,),
            (3,),
            (4,),
            (5,),
            (6,),
            (7,),
            (8,),
            (9,),
            (None,),
        ], "Incorrect result returned"

    def test_numpy_selection(self, duckdb_cursor, integers):
        duckdb_cursor.execute('SELECT * FROM integers')
        duckdb_cursor.execute('SELECT * FROM integers')
        result = duckdb_cursor.fetchnumpy()
        expected = numpy.ma.masked_array(numpy.arange(11), mask=([False] * 10 + [True]))

        numpy.testing.assert_array_equal(result['i'], expected)

    def test_numpy_materialized(self, duckdb_cursor, integers):
        connection = duckdb.connect('')
        cursor = connection.cursor()
        cursor.execute('CREATE TABLE integers (i integer)')
        cursor.execute('INSERT INTO integers VALUES (0),(1),(2),(3),(4),(5),(6),(7),(8),(9),(NULL)')
        rel = connection.table("integers")
        res = rel.aggregate("sum(i)").execute().fetchnumpy()
        assert res['sum(i)'][0] == 45



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import (
    LongType,
    StructType,
    BooleanType,
    StructField,
    StringType,
    IntegerType,
    LongType,
    Row,
    ArrayType,
    MapType,
)
from duckdb.experimental.spark.sql.functions import col, struct, when, lit, array_contains
from duckdb.experimental.spark.sql.functions import sum, avg, max, min, mean, count


@pytest.fixture
def df1(spark):
    data = [("James", 34), ("Michael", 56), ("Robert", 30), ("Maria", 24)]
    dataframe = spark.createDataFrame(data=data, schema=["name", "id"])
    yield dataframe


@pytest.fixture
def df2(spark):
    data2 = [(34, "James"), (45, "Maria"), (45, "Jen"), (34, "Jeff")]
    dataframe = spark.createDataFrame(data=data2, schema=["id", "name"])
    yield dataframe


# https://sparkbyexamples.com/pyspark/pyspark-unionbyname/
@pytest.mark.skip(reason="union_by_name is not supported in the Relation API yet")
class TestDataFrameUnion(object):
    def test_union_by_name(self, df1, df2):
        rel = df1.unionByName(df2)
        res = rel.collect()
        print(res)



import duckdb
import datetime
import numpy as np
import pytest
from conftest import NumpyPandas, ArrowPandas


def create_generic_dataframe(data, pandas):
    return pandas.DataFrame({'col0': pandas.Series(data=data, dtype='object')})


class TestResolveObjectColumns(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_sample_low_correct(self, duckdb_cursor, pandas):
        print(pandas.backend)
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("SET GLOBAL pandas_analyze_sample=3")
        data = [1000008, 6, 9, 4, 1, 6]
        df = create_generic_dataframe(data, pandas)
        roundtripped_df = duckdb.query_df(df, "x", "select * from x", connection=duckdb_conn).df()
        duckdb_df = duckdb_conn.query("select * FROM (VALUES (1000008), (6), (9), (4), (1), (6)) as '0'").df()
        pandas.testing.assert_frame_equal(duckdb_df, roundtripped_df, check_dtype=False)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_sample_low_incorrect_detected(self, duckdb_cursor, pandas):
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("SET GLOBAL pandas_analyze_sample=2")
        # size of list (6) divided by 'pandas_analyze_sample' (2) is the increment used
        # in this case index 0 (1000008) and index 3 ([4]) are checked, which dont match
        data = [1000008, 6, 9, [4], 1, 6]
        df = create_generic_dataframe(data, pandas)
        roundtripped_df = duckdb.query_df(df, "x", "select * from x", connection=duckdb_conn).df()
        # Sample high enough to detect mismatch in types, fallback to VARCHAR
        assert roundtripped_df['col0'].dtype == np.dtype('object')

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_sample_zero(self, duckdb_cursor, pandas):
        duckdb_conn = duckdb.connect()
        # Disable dataframe analyze
        duckdb_conn.execute("SET GLOBAL pandas_analyze_sample=0")
        data = [1000008, 6, 9, 3, 1, 6]
        df = create_generic_dataframe(data, pandas)
        roundtripped_df = duckdb.query_df(df, "x", "select * from x", connection=duckdb_conn).df()
        # Always converts to VARCHAR
        if pandas.backend == 'pyarrow':
            assert roundtripped_df['col0'].dtype == np.dtype('int64')
        else:
            assert roundtripped_df['col0'].dtype == np.dtype('object')

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_sample_low_incorrect_undetected(self, duckdb_cursor, pandas):
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("SET GLOBAL pandas_analyze_sample=1")
        data = [1000008, 6, 9, [4], [1], 6]
        df = create_generic_dataframe(data, pandas)
        # Sample size is too low to detect the mismatch, exception is raised when trying to convert
        with pytest.raises(duckdb.InvalidInputException, match="Failed to cast value: Unimplemented type for cast"):
            roundtripped_df = duckdb.query_df(df, "x", "select * from x", connection=duckdb_conn).df()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_10750(self, duckdb_cursor, pandas):
        max_row_number = 2000
        data = {'id': [i for i in range(max_row_number + 1)], 'content': [None for _ in range(max_row_number + 1)]}

        pdf = pandas.DataFrame(data=data)
        duckdb_cursor.register("content", pdf)
        res = duckdb_cursor.query("select id from content").fetchall()
        expected = [(i,) for i in range(2001)]
        assert res == expected



import duckdb
import pandas as pd
from pytest import mark


@mark.parametrize('timezone', ['UTC', 'CET', 'Asia/Kathmandu'])
def run_pandas_with_tz(timezone):
    con = duckdb.connect()
    con.execute("SET TimeZone = '" + timezone + "'")
    df = pd.DataFrame({"timestamp": [pd.Timestamp("2022-01-01 10:15", tz=timezone)]})
    duck_df = con.from_df(df).df()
    print(df['timestamp'].dtype)
    print(duck_df['timestamp'].dtype)
    print(df)
    print(duck_df)
    assert df.equals(duck_df)



import duckdb
import os
from pytest import raises, mark
import pytest
from conftest import NumpyPandas, ArrowPandas
import datetime

# We only run this test if this env var is set
# FIXME: we can add a custom command line argument to pytest to provide an extension directory
# We can use that instead of checking this environment variable inside of conftest.py's 'require' method
pytestmark = mark.skipif(
    not os.getenv('DUCKDB_PYTHON_TEST_EXTENSION_REQUIRED', False),
    reason='DUCKDB_PYTHON_TEST_EXTENSION_REQUIRED is not set',
)


class TestHTTPFS(object):
    def test_read_json_httpfs(self, require):
        connection = require('httpfs')
        try:
            res = connection.read_json('https://jsonplaceholder.typicode.com/todos')
            assert len(res.types) == 4
        except duckdb.Error as e:
            if '403' in e:
                pytest.skip(reason="Test is flaky, sometimes returns 403")
            else:
                pytest.fail(str(e))

    def test_s3fs(self, require):
        connection = require('httpfs')

        rel = connection.read_csv(f"s3://duckdb-blobs/data/Star_Trek-Season_1.csv", header=True)
        res = rel.fetchone()
        assert res == (1, 0, datetime.date(1965, 2, 28), 0, 0, 0, 1, 0, 1, 1, 1, 0, 0, 6, 0, 0, 0, 0)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_httpfs(self, require, pandas):
        connection = require('httpfs')
        try:
            connection.execute(
                "SELECT id, first_name, last_name FROM PARQUET_SCAN('https://raw.githubusercontent.com/duckdb/duckdb/main/data/parquet-testing/userdata1.parquet') LIMIT 3;"
            )
        except RuntimeError as e:
            # Test will ignore result if it fails due to networking issues while running the test.
            if str(e).startswith("HTTP HEAD error"):
                return
            elif str(e).startswith("Unable to connect"):
                return
            else:
                raise e

        result_df = connection.fetchdf()
        exp_result = pandas.DataFrame(
            {
                'id': pandas.Series([1, 2, 3], dtype="int32"),
                'first_name': ['Amanda', 'Albert', 'Evelyn'],
                'last_name': ['Jordan', 'Freeman', 'Morgan'],
            }
        )
        pandas.testing.assert_frame_equal(result_df, exp_result)

    def test_http_exception(self, require):
        connection = require('httpfs')

        # Read from a bogus HTTPS url, assert that it errors with a non-succesfull status code
        with raises(duckdb.HTTPException) as exc:
            connection.execute("SELECT * FROM PARQUET_SCAN('https://example.com/userdata1.parquet')")

        value = exc.value
        assert value.status_code != 200
        assert value.body == ''
        assert 'Content-Length' in value.headers



import duckdb
import tempfile
import os
import pandas._testing as tm
import datetime
import csv
import pytest
from conftest import NumpyPandas, ArrowPandas, getTimeSeriesData


class TestToCSV(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_basic_to_csv(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': [5, 3, 23, 2], 'b': [45, 234, 234, 2]})
        rel = duckdb.from_df(df)

        rel.to_csv(temp_file_name)

        csv_rel = duckdb.read_csv(temp_file_name)
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_sep(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': [5, 3, 23, 2], 'b': [45, 234, 234, 2]})
        rel = duckdb.from_df(df)

        rel.to_csv(temp_file_name, sep=',')

        csv_rel = duckdb.read_csv(temp_file_name, sep=',')
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_na_rep(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': [5, None, 23, 2], 'b': [45, 234, 234, 2]})
        rel = duckdb.from_df(df)

        rel.to_csv(temp_file_name, na_rep="test")

        csv_rel = duckdb.read_csv(temp_file_name, na_values="test")
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_header(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': [5, None, 23, 2], 'b': [45, 234, 234, 2]})
        rel = duckdb.from_df(df)

        rel.to_csv(temp_file_name)

        csv_rel = duckdb.read_csv(temp_file_name)
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_quotechar(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': ["\'a,b,c\'", None, "hello", "bye"], 'b': [45, 234, 234, 2]})
        rel = duckdb.from_df(df)

        rel.to_csv(temp_file_name, quotechar='\'', sep=',')

        csv_rel = duckdb.read_csv(temp_file_name, sep=',', quotechar='\'')
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_escapechar(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame(
            {
                "c_bool": [True, False],
                "c_float": [1.0, 3.2],
                "c_int": [42, None],
                "c_string": ["a", "b,c"],
            }
        )
        rel = duckdb.from_df(df)
        rel.to_csv(temp_file_name, quotechar='"', escapechar='!')
        csv_rel = duckdb.read_csv(temp_file_name, quotechar='"', escapechar='!')
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_date_format(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame(getTimeSeriesData())
        dt_index = df.index
        df = pandas.DataFrame({"A": dt_index, "B": dt_index.shift(1)}, index=dt_index)
        rel = duckdb.from_df(df)
        rel.to_csv(temp_file_name, date_format="%Y%m%d")

        csv_rel = duckdb.read_csv(temp_file_name, date_format="%Y%m%d")

        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_timestamp_format(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        data = [datetime.time(hour=23, minute=1, second=34, microsecond=234345)]
        df = pandas.DataFrame({'0': pandas.Series(data=data, dtype='object')})
        rel = duckdb.from_df(df)
        rel.to_csv(temp_file_name, timestamp_format='%m/%d/%Y')

        csv_rel = duckdb.read_csv(temp_file_name, timestamp_format='%m/%d/%Y')

        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_quoting_off(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': ['string1', 'string2', 'string3']})
        rel = duckdb.from_df(df)
        rel.to_csv(temp_file_name, quoting=None)

        csv_rel = duckdb.read_csv(temp_file_name)
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_quoting_on(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': ['string1', 'string2', 'string3']})
        rel = duckdb.from_df(df)
        rel.to_csv(temp_file_name, quoting="force")

        csv_rel = duckdb.read_csv(temp_file_name)
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_quoting_quote_all(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': ['string1', 'string2', 'string3']})
        rel = duckdb.from_df(df)
        rel.to_csv(temp_file_name, quoting=csv.QUOTE_ALL)

        csv_rel = duckdb.read_csv(temp_file_name)
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_encoding_incorrect(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': ['string1', 'string2', 'string3']})
        rel = duckdb.from_df(df)
        with pytest.raises(
            duckdb.InvalidInputException, match="Invalid Input Error: The only supported encoding option is 'UTF8"
        ):
            rel.to_csv(temp_file_name, encoding="nope")

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_to_csv_encoding_correct(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': ['string1', 'string2', 'string3']})
        rel = duckdb.from_df(df)
        rel.to_csv(temp_file_name, encoding="UTF-8")
        csv_rel = duckdb.read_csv(temp_file_name)
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_compression_gzip(self, pandas):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pandas.DataFrame({'a': ['string1', 'string2', 'string3']})
        rel = duckdb.from_df(df)
        rel.to_csv(temp_file_name, compression="gzip")
        csv_rel = duckdb.read_csv(temp_file_name, compression="gzip")
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()



import duckdb
import pandas as pd
import numpy as np
import datetime
import math
from decimal import Decimal
from uuid import UUID
import pytz
import pytest


def replace_with_ndarray(obj):
    if hasattr(obj, '__getitem__'):
        if isinstance(obj, dict):
            for key, value in obj.items():
                obj[key] = replace_with_ndarray(value)
        elif isinstance(obj, list):
            for i, item in enumerate(obj):
                obj[i] = replace_with_ndarray(item)
        return np.array(obj)
    return obj


# we need to write our own equality function that considers nan==nan for testing purposes
def recursive_equality(o1, o2):
    import math

    if type(o1) != type(o2):
        return False
    if type(o1) == float and math.isnan(o1) and math.isnan(o2):
        return True
    if o1 is np.ma.masked and o2 is np.ma.masked:
        return True
    try:
        if len(o1) != len(o2):
            return False
        for i in range(len(o1)):
            if not recursive_equality(o1[i], o2[i]):
                return False
        return True
    except:
        return o1 == o2


# Regenerate the 'all_types' list using:

# def get_all_types():
#    conn = duckdb.connect()
#    rel = conn.sql("""
#        select * EXCLUDE
#            time_tz
#        from test_all_types()
#    """)
#    return rel.columns
# all_types = get_all_types()
# for type in all_types:
# 	print(f'\t"{type}",')
# exit()

all_types = [
    "bool",
    "tinyint",
    "smallint",
    "int",
    "bigint",
    "hugeint",
    "utinyint",
    "usmallint",
    "uint",
    "ubigint",
    "date",
    "time",
    "timestamp",
    "timestamp_s",
    "timestamp_ms",
    "timestamp_ns",
    "timestamp_tz",
    "float",
    "double",
    "dec_4_1",
    "dec_9_4",
    "dec_18_6",
    "dec38_10",
    "uuid",
    "interval",
    "varchar",
    "blob",
    "bit",
    "small_enum",
    "medium_enum",
    "large_enum",
    "int_array",
    "double_array",
    "date_array",
    "timestamp_array",
    "timestamptz_array",
    "varchar_array",
    "nested_int_array",
    "struct",
    "struct_of_arrays",
    "array_of_structs",
    "map",
    "union",
    "fixed_int_array",
    "fixed_varchar_array",
    "fixed_nested_int_array",
    "fixed_nested_varchar_array",
    "fixed_struct_array",
    "struct_of_fixed_array",
    "fixed_array_of_int_list",
    "list_of_fixed_int_array",
]


class TestAllTypes(object):
    @pytest.mark.parametrize('cur_type', all_types)
    def test_fetchall(self, cur_type):
        conn = duckdb.connect()
        conn.execute("SET TimeZone =UTC")
        # We replace these values since the extreme ranges are not supported in native-python.
        replacement_values = {
            'timestamp': "'1990-01-01 00:00:00'::TIMESTAMP",
            'timestamp_s': "'1990-01-01 00:00:00'::TIMESTAMP_S",
            'timestamp_ns': "'1990-01-01 00:00:00'::TIMESTAMP_NS",
            'timestamp_ms': "'1990-01-01 00:00:00'::TIMESTAMP_MS",
            'timestamp_tz': "'1990-01-01 00:00:00Z'::TIMESTAMPTZ",
            'date': "'1990-01-01'::DATE",
            'date_array': "[], ['1970-01-01'::DATE, NULL, '0001-01-01'::DATE, '9999-12-31'::DATE,], [NULL::DATE,]",
            'timestamp_array': "[], ['1970-01-01'::TIMESTAMP, NULL, '0001-01-01'::TIMESTAMP, '9999-12-31 23:59:59.999999'::TIMESTAMP,], [NULL::TIMESTAMP,]",
            'timestamptz_array': "[], ['1970-01-01 00:00:00Z'::TIMESTAMPTZ, NULL, '0001-01-01 00:00:00Z'::TIMESTAMPTZ, '9999-12-31 23:59:59.999999Z'::TIMESTAMPTZ,], [NULL::TIMESTAMPTZ,]",
        }
        adjusted_values = {
            'time': """CASE WHEN "time" = '24:00:00'::TIME THEN '23:59:59.999999'::TIME ELSE "time" END AS "time" """,
            'time_tz': """CASE WHEN time_tz = '24:00:00-1559'::TIMETZ THEN '23:59:59.999999-1559'::TIMETZ ELSE time_tz END AS "time_tz" """,
        }
        min_datetime = datetime.datetime.min
        min_datetime_with_utc = min_datetime.replace(tzinfo=pytz.UTC)
        max_datetime = datetime.datetime.max
        max_datetime_with_utc = max_datetime.replace(tzinfo=pytz.UTC)
        correct_answer_map = {
            'bool': [(False,), (True,), (None,)],
            'tinyint': [(-128,), (127,), (None,)],
            'smallint': [(-32768,), (32767,), (None,)],
            'int': [(-2147483648,), (2147483647,), (None,)],
            'bigint': [(-9223372036854775808,), (9223372036854775807,), (None,)],
            'hugeint': [
                (-170141183460469231731687303715884105728,),
                (170141183460469231731687303715884105727,),
                (None,),
            ],
            'utinyint': [(0,), (255,), (None,)],
            'usmallint': [(0,), (65535,), (None,)],
            'uint': [(0,), (4294967295,), (None,)],
            'ubigint': [(0,), (18446744073709551615,), (None,)],
            'time': [(datetime.time(0, 0),), (datetime.time(23, 59, 59, 999999),), (None,)],
            'float': [(-3.4028234663852886e38,), (3.4028234663852886e38,), (None,)],
            'double': [(-1.7976931348623157e308,), (1.7976931348623157e308,), (None,)],
            'dec_4_1': [(Decimal('-999.9'),), (Decimal('999.9'),), (None,)],
            'dec_9_4': [(Decimal('-99999.9999'),), (Decimal('99999.9999'),), (None,)],
            'dec_18_6': [(Decimal('-999999999999.999999'),), (Decimal('999999999999.999999'),), (None,)],
            'dec38_10': [
                (Decimal('-9999999999999999999999999999.9999999999'),),
                (Decimal('9999999999999999999999999999.9999999999'),),
                (None,),
            ],
            'uuid': [
                (UUID('00000000-0000-0000-0000-000000000000'),),
                (UUID('ffffffff-ffff-ffff-ffff-ffffffffffff'),),
                (None,),
            ],
            'varchar': [('🦆🦆🦆🦆🦆🦆',), ('goo\0se',), (None,)],
            'json': [('🦆🦆🦆🦆🦆🦆',), ('goose',), (None,)],
            'blob': [(b'thisisalongblob\x00withnullbytes',), (b'\x00\x00\x00a',), (None,)],
            'bit': [('0010001001011100010101011010111',), ('10101',), (None,)],
            'small_enum': [('DUCK_DUCK_ENUM',), ('GOOSE',), (None,)],
            'medium_enum': [('enum_0',), ('enum_299',), (None,)],
            'large_enum': [('enum_0',), ('enum_69999',), (None,)],
            'date_array': [
                (
                    [],
                    [datetime.date(1970, 1, 1), None, datetime.date.min, datetime.date.max],
                    [
                        None,
                    ],
                )
            ],
            'timestamp_array': [
                (
                    [],
                    [datetime.datetime(1970, 1, 1), None, datetime.datetime.min, datetime.datetime.max],
                    [
                        None,
                    ],
                ),
            ],
            'timestamptz_array': [
                (
                    [],
                    [
                        datetime.datetime(1970, 1, 1, tzinfo=pytz.UTC),
                        None,
                        min_datetime_with_utc,
                        max_datetime_with_utc,
                    ],
                    [
                        None,
                    ],
                ),
            ],
            'int_array': [([],), ([42, 999, None, None, -42],), (None,)],
            'varchar_array': [([],), (['🦆🦆🦆🦆🦆🦆', 'goose', None, ''],), (None,)],
            'double_array': [([],), ([42.0, float('nan'), float('inf'), float('-inf'), None, -42.0],), (None,)],
            'nested_int_array': [
                ([],),
                ([[], [42, 999, None, None, -42], None, [], [42, 999, None, None, -42]],),
                (None,),
            ],
            'struct': [({'a': None, 'b': None},), ({'a': 42, 'b': '🦆🦆🦆🦆🦆🦆'},), (None,)],
            'struct_of_arrays': [
                ({'a': None, 'b': None},),
                ({'a': [42, 999, None, None, -42], 'b': ['🦆🦆🦆🦆🦆🦆', 'goose', None, '']},),
                (None,),
            ],
            'array_of_structs': [([],), ([{'a': None, 'b': None}, {'a': 42, 'b': '🦆🦆🦆🦆🦆🦆'}, None],), (None,)],
            'map': [
                ({'key': [], 'value': []},),
                ({'key': ['key1', 'key2'], 'value': ['🦆🦆🦆🦆🦆🦆', 'goose']},),
                (None,),
            ],
            'time_tz': [(datetime.time(0, 0),), (datetime.time(23, 59, 59, 999999),), (None,)],
            'interval': [
                (datetime.timedelta(0),),
                (datetime.timedelta(days=30969, seconds=999, microseconds=999999),),
                (None,),
            ],
            'timestamp': [(datetime.datetime(1990, 1, 1, 0, 0),)],
            'date': [(datetime.date(1990, 1, 1),)],
            'timestamp_s': [(datetime.datetime(1990, 1, 1, 0, 0),)],
            'timestamp_ns': [(datetime.datetime(1990, 1, 1, 0, 0),)],
            'timestamp_ms': [(datetime.datetime(1990, 1, 1, 0, 0),)],
            'timestamp_tz': [(datetime.datetime(1990, 1, 1, 0, 0, tzinfo=pytz.UTC),)],
            'union': [('Frank',), (5,), (None,)],
            'fixed_int_array': [((None, 2, 3),), ((4, 5, 6),), (None,)],
            'fixed_varchar_array': [(('a', None, 'c'),), (('d', 'e', 'f'),), (None,)],
            'fixed_nested_int_array': [
                (((None, 2, 3), None, (None, 2, 3)),),
                (((4, 5, 6), (None, 2, 3), (4, 5, 6)),),
                (None,),
            ],
            'fixed_nested_varchar_array': [
                ((('a', None, 'c'), None, ('a', None, 'c')),),
                ((('d', 'e', 'f'), ('a', None, 'c'), ('d', 'e', 'f')),),
                (None,),
            ],
            'fixed_struct_array': [
                (({'a': None, 'b': None}, {'a': 42, 'b': '🦆🦆🦆🦆🦆🦆'}, {'a': None, 'b': None}),),
                (({'a': 42, 'b': '🦆🦆🦆🦆🦆🦆'}, {'a': None, 'b': None}, {'a': 42, 'b': '🦆🦆🦆🦆🦆🦆'}),),
                (None,),
            ],
            'struct_of_fixed_array': [
                ({'a': (None, 2, 3), 'b': ('a', None, 'c')},),
                ({'a': (4, 5, 6), 'b': ('d', 'e', 'f')},),
                (None,),
            ],
            'fixed_array_of_int_list': [
                (([], [42, 999, None, None, -42], []),),
                (([42, 999, None, None, -42], [], [42, 999, None, None, -42]),),
                (None,),
            ],
            'list_of_fixed_int_array': [
                ([(None, 2, 3), (4, 5, 6), (None, 2, 3)],),
                ([(4, 5, 6), (None, 2, 3), (4, 5, 6)],),
                (None,),
            ],
        }
        if cur_type in replacement_values:
            result = conn.execute("select " + replacement_values[cur_type]).fetchall()
        elif cur_type in adjusted_values:
            result = conn.execute(f'select {adjusted_values[cur_type]} from test_all_types()').fetchall()
        else:
            result = conn.execute(f'select "{cur_type}" from test_all_types()').fetchall()
        correct_result = correct_answer_map[cur_type]
        assert recursive_equality(result, correct_result)

    def test_bytearray_with_nulls(self):
        con = duckdb.connect(database=':memory:')
        con.execute("CREATE TABLE test (content BLOB)")
        want = bytearray([1, 2, 0, 3, 4])
        con.execute("INSERT INTO test VALUES (?)", [want])

        con.execute("SELECT * from test")
        got = bytearray(con.fetchall()[0][0])
        # Don't truncate the array on the nullbyte
        assert want == bytearray(got)

    @pytest.mark.parametrize('cur_type', all_types)
    def test_fetchnumpy(self, cur_type):
        conn = duckdb.connect()

        correct_answer_map = {
            'bool': np.ma.array(
                [False, True, False],
                mask=[0, 0, 1],
            ),
            'tinyint': np.ma.array(
                [-128, 127, -1],
                mask=[0, 0, 1],
                dtype=np.int8,
            ),
            'smallint': np.ma.array(
                [-32768, 32767, -1],
                mask=[0, 0, 1],
                dtype=np.int16,
            ),
            'int': np.ma.array(
                [-2147483648, 2147483647, -1],
                mask=[0, 0, 1],
                dtype=np.int32,
            ),
            'bigint': np.ma.array(
                [-9223372036854775808, 9223372036854775807, -1],
                mask=[0, 0, 1],
                dtype=np.int64,
            ),
            'utinyint': np.ma.array(
                [0, 255, 42],
                mask=[0, 0, 1],
                dtype=np.uint8,
            ),
            'usmallint': np.ma.array(
                [0, 65535, 42],
                mask=[0, 0, 1],
                dtype=np.uint16,
            ),
            'uint': np.ma.array(
                [0, 4294967295, 42],
                mask=[0, 0, 1],
                dtype=np.uint32,
            ),
            'ubigint': np.ma.array(
                [0, 18446744073709551615, 42],
                mask=[0, 0, 1],
                dtype=np.uint64,
            ),
            'float': np.ma.array(
                [-3.4028234663852886e38, 3.4028234663852886e38, 42.0],
                mask=[0, 0, 1],
                dtype=np.float32,
            ),
            'double': np.ma.array(
                [-1.7976931348623157e308, 1.7976931348623157e308, 42.0],
                mask=[0, 0, 1],
                dtype=np.float64,
            ),
            'uuid': np.ma.array(
                [
                    UUID('00000000-0000-0000-0000-000000000000'),
                    UUID('ffffffff-ffff-ffff-ffff-ffffffffffff'),
                    UUID('00000000-0000-0000-0000-000000000042'),
                ],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'varchar': np.ma.array(
                ['🦆🦆🦆🦆🦆🦆', 'goo\0se', "42"],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'json': np.ma.array(
                ['🦆🦆🦆🦆🦆🦆', 'goose', "42"],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'blob': np.ma.array(
                [b'thisisalongblob\x00withnullbytes', b'\x00\x00\x00a', b"42"],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'interval': np.ma.array(
                [
                    np.timedelta64(0),
                    np.timedelta64(2675722599999999000),
                    np.timedelta64(42),
                ],
                mask=[0, 0, 1],
            ),
            # For timestamp_ns, the lowest value is out-of-range for numpy,
            # such that the conversion yields "Not a Time"
            'timestamp_ns': np.ma.array(
                [
                    np.datetime64("NaT"),
                    np.datetime64(9223372036854775806, "ns"),
                    np.datetime64("1990-01-01T00:42"),
                ],
                mask=[0, 0, 1],
            ),
            # Enums don't have a numpy equivalent and yield pandas Categorical.
            'small_enum': pd.Categorical(
                ['DUCK_DUCK_ENUM', 'GOOSE', np.NaN],
                ordered=True,
            ),
            'medium_enum': pd.Categorical(
                ['enum_0', 'enum_299', np.NaN],
                ordered=True,
            ),
            'large_enum': pd.Categorical(
                ['enum_0', 'enum_69999', np.NaN],
                ordered=True,
            ),
            # The following types don't have a numpy equivalent and yield
            # object arrays:
            'int_array': np.ma.array(
                [
                    [],
                    [42, 999, None, None, -42],
                    None,
                ],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'varchar_array': np.ma.array(
                [
                    [],
                    ['🦆🦆🦆🦆🦆🦆', 'goose', None, ''],
                    None,
                ],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'double_array': np.ma.array(
                [
                    [],
                    [42.0, float('nan'), float('inf'), float('-inf'), None, -42.0],
                    None,
                ],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'nested_int_array': np.ma.array(
                [
                    [],
                    [[], [42, 999, None, None, -42], None, [], [42, 999, None, None, -42]],
                    None,
                ],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'struct': np.ma.array(
                [
                    {'a': None, 'b': None},
                    {'a': 42, 'b': '🦆🦆🦆🦆🦆🦆'},
                    None,
                ],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'struct_of_arrays': np.ma.array(
                [
                    {'a': None, 'b': None},
                    {'a': [42, 999, None, None, -42], 'b': ['🦆🦆🦆🦆🦆🦆', 'goose', None, '']},
                    None,
                ],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'array_of_structs': np.ma.array(
                [
                    [],
                    [{'a': None, 'b': None}, {'a': 42, 'b': '🦆🦆🦆🦆🦆🦆'}, None],
                    None,
                ],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'map': np.ma.array(
                [
                    {'key': [], 'value': []},
                    {'key': ['key1', 'key2'], 'value': ['🦆🦆🦆🦆🦆🦆', 'goose']},
                    None,
                ],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'time': np.ma.array(
                ['00:00:00', '24:00:00', None],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'time_tz': np.ma.array(
                ['00:00:00', '23:59:59.999999', None],
                mask=[0, 0, 1],
                dtype=object,
            ),
            'union': np.ma.array(['Frank', 5, None], mask=[0, 0, 1], dtype=object),
        }
        correct_answer_map = replace_with_ndarray(correct_answer_map)

        # The following types don't have a numpy equivalent, and are coerced to
        # floating point types by fetchnumpy():
        # - 'uhugeint'
        # - 'hugeint'
        # - 'dec_4_1'
        # - 'dec_9_4'
        # - 'dec_18_6'
        # - 'dec38_10'

        # The following types lead to errors:
        # Conversion Error: Could not convert DATE to nanoseconds
        # - 'date'
        # Conversion Error: Date out of range in timestamp conversion
        # - 'timestamp_array'
        # - 'timestamptz_array'
        # Conversion Error: Could not convert Timestamp(US) to Timestamp(NS)
        # - 'timestamp'
        # - 'timestamp_s'
        # - 'timestamp_ms'
        # - 'timestamp_tz'
        # SystemError: <built-in function __import__> returned a result with an exception set
        # - 'date_array'

        rel = conn.table_function("test_all_types")
        if cur_type not in correct_answer_map:
            return
        result = rel.project(f'"{cur_type}"').fetchnumpy()
        result = result[cur_type]
        correct_answer = correct_answer_map[cur_type]
        if isinstance(result, pd.Categorical) or result.dtype == object:
            assert recursive_equality(list(result), list(correct_answer))
        else:
            # assert_equal compares NaN equal, but also compares masked
            # elements equal to any unmasked element
            if isinstance(result, np.ma.MaskedArray) or isinstance(correct_answer, np.ma.MaskedArray):
                assert np.all(result.mask == correct_answer.mask)
            np.testing.assert_equal(result, correct_answer)

    @pytest.mark.parametrize('cur_type', all_types)
    def test_arrow(self, cur_type):
        try:
            import pyarrow as pa
        except:
            return
        # We skip those since the extreme ranges are not supported in arrow.
        replacement_values = {'interval': "INTERVAL '2 years'"}
        # We do not round trip enum types
        enum_types = {'small_enum', 'medium_enum', 'large_enum', 'double_array'}

        # uhugeint currently not supported by arrow
        skip_types = {'uhugeint'}
        if cur_type in skip_types:
            return

        conn = duckdb.connect()
        if cur_type in replacement_values:
            arrow_table = conn.execute("select " + replacement_values[cur_type]).arrow()
        else:
            arrow_table = conn.execute(f'select "{cur_type}" from test_all_types()').arrow()
        if cur_type in enum_types:
            round_trip_arrow_table = conn.execute("select * from arrow_table").arrow()
            result_arrow = conn.execute("select * from arrow_table").fetchall()
            result_roundtrip = conn.execute("select * from round_trip_arrow_table").fetchall()
            assert recursive_equality(result_arrow, result_roundtrip)
        else:
            round_trip_arrow_table = conn.execute("select * from arrow_table").arrow()
            assert arrow_table.equals(round_trip_arrow_table, check_metadata=True)

    @pytest.mark.parametrize('cur_type', all_types)
    def test_pandas(self, cur_type):
        # We skip those since the extreme ranges are not supported in python.
        replacement_values = {
            'timestamp': "'1990-01-01 00:00:00'::TIMESTAMP",
            'timestamp_s': "'1990-01-01 00:00:00'::TIMESTAMP_S",
            'timestamp_ns': "'1990-01-01 00:00:00'::TIMESTAMP_NS",
            'timestamp_ms': "'1990-01-01 00:00:00'::TIMESTAMP_MS",
            'timestamp_tz': "'1990-01-01 00:00:00Z'::TIMESTAMPTZ",
            'date': "'1990-01-01'::DATE",
            'date_array': "[], ['1970-01-01'::DATE, NULL, '0001-01-01'::DATE, '9999-12-31'::DATE,], [NULL::DATE,]",
            'timestamp_array': "[], ['1970-01-01'::TIMESTAMP, NULL, '0001-01-01'::TIMESTAMP, '9999-12-31 23:59:59.999999'::TIMESTAMP,], [NULL::TIMESTAMP,]",
            'timestamptz_array': "[], ['1970-01-01 00:00:00Z'::TIMESTAMPTZ, NULL, '0001-01-01 00:00:00Z'::TIMESTAMPTZ, '9999-12-31 23:59:59.999999Z'::TIMESTAMPTZ,], [NULL::TIMESTAMPTZ,]",
        }

        adjusted_values = {
            'time': """CASE WHEN "time" = '24:00:00'::TIME THEN '23:59:59.999999'::TIME ELSE "time" END AS "time" """,
        }
        conn = duckdb.connect()
        conn.execute("SET timezone = UTC")
        if cur_type in replacement_values:
            dataframe = conn.execute("select " + replacement_values[cur_type]).df()
        elif cur_type in adjusted_values:
            dataframe = conn.execute(f'select {adjusted_values[cur_type]} from test_all_types()').df()
        else:
            dataframe = conn.execute(f'select "{cur_type}" from test_all_types()').df()
        print(cur_type)
        round_trip_dataframe = conn.execute("select * from dataframe").df()
        result_dataframe = conn.execute("select * from dataframe").fetchall()
        print(round_trip_dataframe)
        result_roundtrip = conn.execute("select * from round_trip_dataframe").fetchall()
        assert recursive_equality(result_dataframe, result_roundtrip)



import duckdb
import pandas as pd
import numpy
import datetime
import time


class TestPartitionedPandasScan(object):
    def test_parallel_pandas(self, duckdb_cursor):
        con = duckdb.connect()
        df = pd.DataFrame({'i': numpy.arange(10000000)})

        con.register('df', df)

        seq_results = con.execute("SELECT SUM(i) FROM df").fetchall()

        con.execute("PRAGMA threads=4")
        parallel_results = con.execute("SELECT SUM(i) FROM df").fetchall()

        assert seq_results[0][0] == 49999995000000
        assert parallel_results[0][0] == 49999995000000



import duckdb
import os
import pytest


def get_tables(con):
    tbls = con.execute("SHOW TABLES").fetchall()
    tbls = [x[0] for x in tbls]
    tbls.sort()
    return tbls


def test_multiple_writes():
    try:
        os.remove("test.db")
    except:
        pass
    con1 = duckdb.connect("test.db")
    con2 = duckdb.connect("test.db")
    con1.execute("CREATE TABLE foo1 as SELECT 1 as a, 2 as b")
    con2.execute("CREATE TABLE bar1 as SELECT 2 as a, 3 as b")
    con2.close()
    con1.close()
    con3 = duckdb.connect("test.db")
    tbls = get_tables(con3)
    assert tbls == ['bar1', 'foo1']
    del con1
    del con2
    del con3

    try:
        os.remove("test.db")
    except:
        pass


def test_multiple_writes_memory():
    con1 = duckdb.connect()
    con2 = duckdb.connect()
    con1.execute("CREATE TABLE foo1 as SELECT 1 as a, 2 as b")
    con2.execute("CREATE TABLE bar1 as SELECT 2 as a, 3 as b")
    con3 = duckdb.connect(":memory:")
    tbls = get_tables(con1)
    assert tbls == ['foo1']
    tbls = get_tables(con2)
    assert tbls == ['bar1']
    tbls = get_tables(con3)
    assert tbls == []
    del con1
    del con2
    del con3


def test_multiple_writes_named_memory():
    con1 = duckdb.connect(":memory:1")
    con2 = duckdb.connect(":memory:1")
    con1.execute("CREATE TABLE foo1 as SELECT 1 as a, 2 as b")
    con2.execute("CREATE TABLE bar1 as SELECT 2 as a, 3 as b")
    con3 = duckdb.connect(":memory:1")
    tbls = get_tables(con3)
    assert tbls == ['bar1', 'foo1']
    del con1
    del con2
    del con3


def test_diff_config():
    con1 = duckdb.connect("test.db", False)
    with pytest.raises(
        duckdb.ConnectionException,
        match="Can't open a connection to same database file with a different configuration than existing connections",
    ):
        con2 = duckdb.connect("test.db", True)
    con1.close()
    del con1



import duckdb
import numpy as np

try:
    import pyarrow as pa

    can_run = True
except:
    can_run = False


def check_equal(duckdb_conn):
    true_result = duckdb_conn.execute("SELECT * from test").fetchall()
    arrow_result = duckdb_conn.execute("SELECT * from testarrow").fetchall()
    assert arrow_result == true_result


def create_and_register_arrow_table(column_list, duckdb_conn):
    pydict = {name: data for (name, _, data) in column_list}
    arrow_schema = pa.schema([(name, dtype) for (name, dtype, _) in column_list])
    res = pa.Table.from_pydict(pydict, schema=arrow_schema)

    duck_from_arrow = duckdb_conn.from_arrow(res)
    duck_from_arrow.create("testarrow")


def create_and_register_comparison_result(column_list, duckdb_conn):
    columns = ",".join([f'{name} {dtype}' for (name, dtype, _) in column_list])
    column_amount = len(column_list)
    assert column_amount
    row_amount = len(column_list[0][2])
    inserted_values = []
    for row in range(row_amount):
        for col in range(column_amount):
            inserted_values.append(column_list[col][2][row])
    inserted_values = tuple(inserted_values)

    column_format = ",".join(['?' for _ in range(column_amount)])
    row_format = ",".join([f"({column_format})" for _ in range(row_amount)])
    query = f"""CREATE TABLE test ({columns});
        INSERT INTO test VALUES {row_format};
    """

    duckdb_conn.execute(query, inserted_values)


class TestArrowListType(object):
    def test_regular_list(self):
        if not can_run:
            return
        duckdb_conn = duckdb.connect()

        n = 5  # Amount of lists
        generated_size = 3  # Size of each list
        list_size = -1  # Argument passed to `pa._list()`

        data = [np.random.random((generated_size)) for _ in range(n)]
        list_type = pa.list_(pa.float32(), list_size=list_size)

        create_and_register_arrow_table(
            [
                ('a', list_type, data),
            ],
            duckdb_conn,
        )
        create_and_register_comparison_result(
            [
                ('a', 'FLOAT[]', data),
            ],
            duckdb_conn,
        )

        check_equal(duckdb_conn)

    def test_fixedsize_list(self):
        if not can_run:
            return
        duckdb_conn = duckdb.connect()

        n = 5  # Amount of lists
        generated_size = 3  # Size of each list
        list_size = 3  # Argument passed to `pa._list()`

        data = [np.random.random((generated_size)) for _ in range(n)]
        list_type = pa.list_(pa.float32(), list_size=list_size)

        create_and_register_arrow_table(
            [
                ('a', list_type, data),
            ],
            duckdb_conn,
        )
        create_and_register_comparison_result(
            [
                ('a', f'FLOAT[{list_size}]', data),
            ],
            duckdb_conn,
        )

        check_equal(duckdb_conn)



import duckdb
import tempfile
import os
import pandas as pd
import tempfile
import pandas._testing as tm
import datetime
import csv
import pytest


class TestToParquet(object):
    def test_basic_to_parquet(self):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pd.DataFrame({'a': [5, 3, 23, 2], 'b': [45, 234, 234, 2]})
        rel = duckdb.from_df(df)

        rel.to_parquet(temp_file_name)

        csv_rel = duckdb.read_parquet(temp_file_name)
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()

    def test_compression_gzip(self):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        df = pd.DataFrame({'a': ['string1', 'string2', 'string3']})
        rel = duckdb.from_df(df)
        rel.to_parquet(temp_file_name, compression="gzip")
        csv_rel = duckdb.read_parquet(temp_file_name, compression="gzip")
        assert rel.execute().fetchall() == csv_rel.execute().fetchall()



import duckdb
import pytest

pyarrow = pytest.importorskip('pyarrow')


@pytest.mark.parametrize('arrow_large_buffer_size', [True, False])
def test_10795(arrow_large_buffer_size):
    conn = duckdb.connect()
    conn.sql(f"set arrow_large_buffer_size={arrow_large_buffer_size}")
    arrow = conn.sql("select map(['non-inlined string', 'test', 'duckdb'], [42, 1337, 123]) as map").to_arrow_table()
    assert arrow.to_pydict() == {'map': [[('non-inlined string', 42), ('test', 1337), ('duckdb', 123)]]}



import duckdb
import os

try:
    import pyarrow as pa
    from pyarrow import parquet as pq
    import numpy as np

    can_run = True
except:
    can_run = False


class TestArrowLargeString(object):
    def test_large_string_type(self, duckdb_cursor):
        if not can_run:
            return

        schema = pa.schema([("data", pa.large_string())])
        inputs = [pa.array(["foo", "baaaar", "b"], type=pa.large_string())]
        arrow_table = pa.Table.from_arrays(inputs, schema=schema)

        rel = duckdb.from_arrow(arrow_table)
        res = rel.execute().fetchall()
        assert res == [('foo',), ('baaaar',), ('b',)]



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import Row
import textwrap


class TestSparkReadCSV(object):
    def test_read_csv(self, spark, tmp_path):
        file_path = tmp_path / 'basic.csv'
        with open(file_path, 'w+') as f:
            f.write(
                textwrap.dedent(
                    """
				1,2
				3,4
				5,6
			"""
                )
            )
        file_path = file_path.as_posix()
        df = spark.read.csv(file_path)
        res = df.collect()
        assert res == [Row(column0=1, column1=2), Row(column0=3, column1=4), Row(column0=5, column1=6)]



import os

script_dir = os.path.dirname(__file__)
from typing import List, Dict, Union
import json

lines: List[str] = [file for file in open(f'{script_dir}/imports.py').read().split('\n') if file != '']


class ImportCacheAttribute:
    def __init__(self, full_path: str):
        parts = full_path.split('.')
        self.type = "attribute"
        self.name = parts[-1]
        self.full_path = full_path
        self.children: Dict[str, "ImportCacheAttribute"] = {}

    def has_item(self, item_name: str) -> bool:
        return item_name in self.children

    def get_item(self, item_name: str) -> "ImportCacheAttribute":
        assert item_name in self.children
        return self.children[item_name]

    def add_item(self, item: "ImportCacheAttribute"):
        assert not self.has_item(item.full_path)
        self.children[item.full_path] = item

    def __repr__(self) -> str:
        return str(self.children)

    def populate_json(self, json_data: dict):
        json_data[self.full_path] = {
            "type": self.type,
            "full_path": self.full_path,
            "name": self.name,
            "children": list(self.children.keys()),
        }
        for item in self.children:
            self.children[item].populate_json(json_data)


class ImportCacheModule:
    def __init__(self, full_path):
        parts = full_path.split('.')
        self.type = "module"
        self.name = parts[-1]
        self.full_path = full_path
        self.items: Dict[str, Union[ImportCacheAttribute, "ImportCacheModule"]] = {}

    def add_item(self, item: Union[ImportCacheAttribute, "ImportCacheModule"]):
        assert self.full_path != item.full_path
        assert not self.has_item(item.full_path)
        self.items[item.full_path] = item

    def get_item(self, item_name: str) -> Union[ImportCacheAttribute, "ImportCacheModule"]:
        assert self.has_item(item_name)
        return self.items[item_name]

    def populate_json(self, json_data: dict):
        json_data[self.full_path] = {
            "type": self.type,
            "full_path": self.full_path,
            "name": self.name,
            "children": list(self.items.keys()),
        }
        for key_name in self.items:
            self.items[key_name].populate_json(json_data)

    def has_item(self, item_name: str) -> bool:
        return item_name in self.items

    def __repr__(self) -> str:
        return str(self.items)

    def root_module(self) -> bool:
        return self.name == self.full_path


class ImportCacheGenerator:
    def __init__(self):
        self.modules: Dict[str, ImportCacheModule] = {}

    def add_module(self, path: str):
        assert path.startswith('import')
        path = path[7:]
        module = ImportCacheModule(path)
        self.modules[module.full_path] = module

        # Add it to the parent module if present
        parts = path.split('.')
        if len(parts) == 1:
            return

        # This works back from the furthest child module to the top level module
        child_module = module
        for i in range(1, len(parts)):
            parent_path = '.'.join(parts[: len(parts) - i])
            parent_module = self.add_or_get_module(parent_path)
            parent_module.add_item(child_module)
            child_module = parent_module

    def add_or_get_module(self, module_name: str) -> ImportCacheModule:
        if module_name not in self.modules:
            self.add_module(f'import {module_name}')
        return self.get_module(module_name)

    def get_module(self, module_name: str) -> ImportCacheModule:
        if module_name not in self.modules:
            raise ValueError("Import the module before registering its attributes!")
        return self.modules[module_name]

    def get_item(self, item_name: str) -> Union[ImportCacheModule, ImportCacheAttribute]:
        parts = item_name.split('.')
        if len(parts) == 1:
            return self.get_module(item_name)

        parent = self.get_module(parts[0])
        for i in range(1, len(parts)):
            child_path = '.'.join(parts[: i + 1])
            if parent.has_item(child_path):
                parent = parent.get_item(child_path)
            else:
                attribute = ImportCacheAttribute(child_path)
                parent.add_item(attribute)
                parent = attribute
        return parent

    def add_attribute(self, path: str):
        assert not path.startswith('import')
        parts = path.split('.')
        assert len(parts) >= 2
        self.get_item(path)

    def populate_json(self, json_data: dict):
        for module_name in self.modules:
            self.modules[module_name].populate_json(json_data)

    def to_json(self):
        json_data = {}
        self.populate_json(json_data)
        return json_data


generator = ImportCacheGenerator()

for line in lines:
    if line.startswith('#'):
        continue
    if line.startswith('import'):
        generator.add_module(line)
    else:
        generator.add_attribute(line)

# Load existing JSON data from a file if it exists
existing_json_data = {}
json_cache_path = os.path.join(script_dir, "cache_data.json")
try:
    with open(json_cache_path, "r") as file:
        existing_json_data = json.load(file)
except FileNotFoundError:
    pass


def update_json(existing: dict, new: dict):
    for item in new:
        if item not in existing:
            continue
        object = new[item]
        if isinstance(object, dict):
            object.update(existing[item])
            update_json(object, existing[item])


# Merge the existing JSON data with the new data
json_data = generator.to_json()
update_json(existing_json_data, json_data)

# Save the merged JSON data back to the file
with open(json_cache_path, "w") as file:
    json.dump(json_data, file, indent=4)



import duckdb
import pandas as pd
import numpy


class TestPandasString(object):
    def test_pandas_string(self, duckdb_cursor):
        strings = numpy.array(['foo', 'bar', 'baz'])

        # https://pandas.pydata.org/pandas-docs/stable/user_guide/text.html
        df_in = pd.DataFrame(
            {
                'object': pd.Series(strings, dtype='object'),
            }
        )
        # Only available in pandas 1.0.0
        if hasattr(pd, 'StringDtype'):
            df_in['string'] = pd.Series(strings, dtype=pd.StringDtype())

        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()

        assert numpy.all(df_out['object'] == strings)
        if hasattr(pd, 'StringDtype'):
            assert numpy.all(df_out['string'] == strings)

    def test_bug_2467(self, duckdb_cursor):
        N = 1_000_000
        # Create DataFrame with string attribute
        df = pd.DataFrame({"city": ["Amsterdam", "New York", "London"] * N})
        # Copy Dataframe to DuckDB
        con = duckdb.connect()
        con.register("df", df)
        con.execute(
            f"""
            CREATE TABLE t1 AS SELECT * FROM df
        """
        )
        assert (
            con.execute(
                f"""
            SELECT count(*) from t1
        """
            ).fetchall()
            == [(3000000,)]
        )



import duckdb
import pytest

try:
    import pyarrow as pa

    can_run = True
except:
    can_run = False
from conftest import NumpyPandas, ArrowPandas


class Test3654(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_3654_pandas(self, duckdb_cursor, pandas):
        df1 = pandas.DataFrame(
            {
                'id': [1, 1, 2],
            }
        )
        con = duckdb.connect()
        con.register("df1", df1)
        rel = con.view("df1")
        print(rel.execute().fetchall())
        assert rel.execute().fetchall() == [(1,), (1,), (2,)]

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_3654_arrow(self, duckdb_cursor, pandas):
        if not can_run:
            return

        df1 = pandas.DataFrame(
            {
                'id': [1, 1, 2],
            }
        )
        table = pa.Table.from_pandas(df1)
        con = duckdb.connect()
        con.register("df1", table)
        rel = con.view("df1")
        print(rel.execute().fetchall())
        assert rel.execute().fetchall() == [(1,), (1,), (2,)]



import duckdb
import os
import datetime
import pytest

try:
    import pyarrow as pa
    import pandas as pd

    can_run = True
except:
    can_run = False


class TestArrowTime(object):
    def test_time_types(self, duckdb_cursor):
        if not can_run:
            return

        data = (
            pa.array([1], type=pa.time32('s')),
            pa.array([1000], type=pa.time32('ms')),
            pa.array([1000000], pa.time64('us')),
            pa.array([1000000000], pa.time64('ns')),
        )
        arrow_table = pa.Table.from_arrays([data[0], data[1], data[2], data[3]], ['a', 'b', 'c', 'd'])
        rel = duckdb.from_arrow(arrow_table).arrow()
        assert rel['a'] == arrow_table['c']
        assert rel['b'] == arrow_table['c']
        assert rel['c'] == arrow_table['c']
        assert rel['d'] == arrow_table['c']

    def test_time_null(self, duckdb_cursor):
        if not can_run:
            return
        data = (
            pa.array([None], type=pa.time32('s')),
            pa.array([None], type=pa.time32('ms')),
            pa.array([None], pa.time64('us')),
            pa.array([None], pa.time64('ns')),
        )
        arrow_table = pa.Table.from_arrays([data[0], data[1], data[2], data[3]], ['a', 'b', 'c', 'd'])
        rel = duckdb.from_arrow(arrow_table).arrow()
        assert rel['a'] == arrow_table['c']
        assert rel['b'] == arrow_table['c']
        assert rel['c'] == arrow_table['c']
        assert rel['d'] == arrow_table['c']

    def test_max_times(self, duckdb_cursor):
        if not can_run:
            return
        data = pa.array([2147483647000000], type=pa.time64('us'))
        result = pa.Table.from_arrays([data], ['a'])
        # Max Sec
        data = pa.array([2147483647], type=pa.time32('s'))
        arrow_table = pa.Table.from_arrays([data], ['a'])
        rel = duckdb.from_arrow(arrow_table).arrow()
        assert rel['a'] == result['a']

        # Max MSec
        data = pa.array([2147483647000], type=pa.time64('us'))
        result = pa.Table.from_arrays([data], ['a'])
        data = pa.array([2147483647], type=pa.time32('ms'))
        arrow_table = pa.Table.from_arrays([data], ['a'])
        rel = duckdb.from_arrow(arrow_table).arrow()
        assert rel['a'] == result['a']

        # Max NSec
        data = pa.array([9223372036854774], type=pa.time64('us'))
        result = pa.Table.from_arrays([data], ['a'])
        data = pa.array([9223372036854774000], type=pa.time64('ns'))
        arrow_table = pa.Table.from_arrays([data], ['a'])
        rel = duckdb.from_arrow(arrow_table).arrow()

        print(rel['a'])
        print(result['a'])
        assert rel['a'] == result['a']



import duckdb
import tempfile
import os
import pandas as pd
import tempfile
import pandas._testing as tm
import datetime
import csv
import pytest


class TestGetAttribute(object):
    def test_basic_getattr(self, duckdb_cursor):
        rel = duckdb_cursor.sql('select i as a, (i + 5) % 10 as b, (i + 2) % 3 as c from range(100) tbl(i)')
        assert rel.a.fetchmany(5) == [(0,), (1,), (2,), (3,), (4,)]
        assert rel.b.fetchmany(5) == [(5,), (6,), (7,), (8,), (9,)]
        assert rel.c.fetchmany(5) == [(2,), (0,), (1,), (2,), (0,)]

    def test_basic_getitem(self, duckdb_cursor):
        rel = duckdb_cursor.sql('select i as a, (i + 5) % 10 as b, (i + 2) % 3 as c from range(100) tbl(i)')
        assert rel['a'].fetchmany(5) == [(0,), (1,), (2,), (3,), (4,)]
        assert rel['b'].fetchmany(5) == [(5,), (6,), (7,), (8,), (9,)]
        assert rel['c'].fetchmany(5) == [(2,), (0,), (1,), (2,), (0,)]

    def test_getitem_nonexistant(self, duckdb_cursor):
        rel = duckdb_cursor.sql('select i as a, (i + 5) % 10 as b, (i + 2) % 3 as c from range(100) tbl(i)')
        with pytest.raises(AttributeError):
            rel['d']

    def test_getattr_nonexistant(self, duckdb_cursor):
        rel = duckdb_cursor.sql('select i as a, (i + 5) % 10 as b, (i + 2) % 3 as c from range(100) tbl(i)')
        with pytest.raises(AttributeError):
            rel.d

    def test_getattr_collision(self, duckdb_cursor):
        rel = duckdb_cursor.sql('select i as df from range(100) tbl(i)')

        # 'df' also exists as a method on DuckDBPyRelation
        assert rel.df.__class__ != duckdb.DuckDBPyRelation

    def test_getitem_collision(self, duckdb_cursor):
        rel = duckdb_cursor.sql('select i as df from range(100) tbl(i)')

        # this case is not an issue on __getitem__
        assert rel['df'].__class__ == duckdb.DuckDBPyRelation

    def test_getitem_struct(self, duckdb_cursor):
        rel = duckdb_cursor.sql("select {'a':5, 'b':6} as a, 5 as b")
        assert rel['a']['a'].fetchall()[0][0] == 5
        assert rel['a']['b'].fetchall()[0][0] == 6

    def test_getattr_struct(self, duckdb_cursor):
        rel = duckdb_cursor.sql("select {'a':5, 'b':6} as a, 5 as b")
        assert rel.a.a.fetchall()[0][0] == 5
        assert rel.a.b.fetchall()[0][0] == 6

    def test_getattr_spaces(self, duckdb_cursor):
        rel = duckdb_cursor.sql('select 42 as "hello world"')
        assert rel['hello world'].fetchall()[0][0] == 42



import pytest
import duckdb
import numpy as np
import sys
from conftest import NumpyPandas, ArrowPandas


def assert_create(internal_data, expected_result, data_type, pandas):
    conn = duckdb.connect()
    df_in = pandas.DataFrame(data=internal_data, dtype=data_type)

    conn.execute("CREATE TABLE t AS SELECT * FROM df_in")

    result = conn.execute("SELECT * FROM t").fetchall()
    assert result == expected_result


def assert_create_register(internal_data, expected_result, data_type, pandas):
    conn = duckdb.connect()
    df_in = pandas.DataFrame(data=internal_data, dtype=data_type)
    conn.register("dataframe", df_in)
    conn.execute("CREATE TABLE t AS SELECT * FROM dataframe")

    result = conn.execute("SELECT * FROM t").fetchall()
    assert result == expected_result


class TestCreateTableFromPandas(object):
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_integer_create_table(self, duckdb_cursor, pandas):
        if sys.version_info.major < 3:
            return
        # FIXME: This should work with other data types e.g., int8...
        data_types = ['Int8', 'Int16', 'Int32', 'Int64']
        internal_data = [1, 2, 3, 4]
        expected_result = [(1,), (2,), (3,), (4,)]
        for data_type in data_types:
            print(data_type)
            assert_create_register(internal_data, expected_result, data_type, pandas)
            assert_create(internal_data, expected_result, data_type, pandas)

    # FIXME: Also test other data types



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import (
    LongType,
    StructType,
    BooleanType,
    StructField,
    StringType,
    IntegerType,
    LongType,
    Row,
    ArrayType,
    MapType,
)
from duckdb.experimental.spark.sql.functions import col, struct, when, lit, array_contains
import duckdb
import re


class TestDataFrameFilter(object):
    def test_dataframe_filter(self, spark):
        data = [
            (("James", "", "Smith"), ["Java", "Scala", "C++"], "OH", "M"),
            (("Anna", "Rose", ""), ["Spark", "Java", "C++"], "CA", "F"),
            (("Julia", "", "Williams"), ["CSharp", "VB"], "OH", "F"),
            (("Maria", "Anne", "Jones"), ["CSharp", "VB"], "NY", "M"),
            (("Jen", "Mary", "Brown"), ["CSharp", "VB"], "NY", "M"),
            (("Mike", "Mary", "Williams"), ["Python", "VB"], "OH", "M"),
        ]

        schema = StructType(
            [
                StructField(
                    'name',
                    StructType(
                        [
                            StructField('firstname', StringType(), True),
                            StructField('middlename', StringType(), True),
                            StructField('lastname', StringType(), True),
                        ]
                    ),
                ),
                StructField('languages', ArrayType(StringType()), True),
                StructField('state', StringType(), True),
                StructField('gender', StringType(), True),
            ]
        )

        df = spark.createDataFrame(data=data, schema=schema)

        # --- Tests ---

        # Using equals condition
        df2 = df.filter(df.state == "OH")
        res = df2.collect()
        assert res[0].state == 'OH'

        # not equals condition
        df2 = df.filter(df.state != "OH")
        df2 = df.filter(~(df.state == "OH"))
        res = df2.collect()
        for item in res:
            assert item.state == 'NY' or item.state == 'CA'

        df2 = df.filter(col("state") == "OH")
        res = df2.collect()
        assert res[0].state == 'OH'

        df2 = df.filter("gender == 'M'")
        res = df2.collect()
        assert res[0].gender == 'M'

        df2 = df.filter("gender != 'M'")
        res = df2.collect()
        assert res[0].gender == 'F'

        df2 = df.filter("gender <> 'M'")
        res = df2.collect()
        assert res[0].gender == 'F'

        # Filter multiple condition
        df2 = df.filter((df.state == "OH") & (df.gender == "M"))
        res = df2.collect()
        assert len(res) == 2
        for item in res:
            assert item.gender == 'M' and item.state == 'OH'

        # Filter IS IN List values
        li = ["OH", "NY"]
        df2 = df.filter(df.state.isin(li))
        res = df2.collect()
        for item in res:
            assert item.state == 'OH' or item.state == 'NY'

        # Filter NOT IS IN List values
        # These show all records with NY (NY is not part of the list)
        df2 = df.filter(~df.state.isin(li))
        res = df2.collect()
        for item in res:
            assert item.state != 'OH' and item.state != 'NY'

        df2 = df.filter(df.state.isin(li) == False)
        res2 = df2.collect()
        assert res2 == res

        # Using startswith
        df2 = df.filter(df.state.startswith("N"))
        res = df2.collect()
        for item in res:
            assert item.state == 'NY'

        # using endswith
        df2 = df.filter(df.state.endswith("H"))
        res = df2.collect()
        for item in res:
            assert item.state == 'OH'

        # contains
        df2 = df.filter(df.state.contains("H"))
        res = df2.collect()
        for item in res:
            assert item.state == 'OH'

        data2 = [(2, "Michael Rose"), (3, "Robert Williams"), (4, "Rames Rose"), (5, "Rames rose")]
        df2 = spark.createDataFrame(data=data2, schema=["id", "name"])

        # like - SQL LIKE pattern
        df3 = df2.filter(df2.name.like("%rose%"))
        res = df3.collect()
        assert res == [Row(id=5, name='Rames rose')]

        # rlike - SQL RLIKE pattern (LIKE with Regex)
        # This check case insensitive
        df3 = df2.filter(df2.name.rlike("(?i)^*rose$"))
        res = df3.collect()
        assert res == [Row(id=2, name='Michael Rose'), Row(id=4, name='Rames Rose'), Row(id=5, name='Rames rose')]

        df2 = df.filter(array_contains(df.languages, "Java"))
        res = df2.collect()
        assert res == [
            Row(
                name={'firstname': 'James', 'middlename': '', 'lastname': 'Smith'},
                languages=['Java', 'Scala', 'C++'],
                state='OH',
                gender='M',
            ),
            Row(
                name={'firstname': 'Anna', 'middlename': 'Rose', 'lastname': ''},
                languages=['Spark', 'Java', 'C++'],
                state='CA',
                gender='F',
            ),
        ]

        df2 = df.filter(df.name.lastname == "Williams")
        res = df2.collect()
        assert res == [
            Row(
                name={'firstname': 'Julia', 'middlename': '', 'lastname': 'Williams'},
                languages=['CSharp', 'VB'],
                state='OH',
                gender='F',
            ),
            Row(
                name={'firstname': 'Mike', 'middlename': 'Mary', 'lastname': 'Williams'},
                languages=['Python', 'VB'],
                state='OH',
                gender='M',
            ),
        ]



import pandas as pd
import pytest
import duckdb
import sys
import gc


class TestPandasDFNone(object):
    # This used to decrease the ref count of None
    def test_none_deref(self):
        con = duckdb.connect()
        df = con.sql("select NULL::VARCHAR as a from range(1000000)").df()



import duckdb
import pandas as pd
import numpy
import string
from packaging import version


def round_trip(data, pandas_type):
    df_in = pd.DataFrame(
        {
            'object': pd.Series(data, dtype=pandas_type),
        }
    )

    df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()
    print(df_out)
    print(df_in)
    assert df_out.equals(df_in)


class TestNumpyNullableTypes(object):
    def test_pandas_numeric(self):
        base_df = pd.DataFrame({'a': range(10)})

        data_types = [
            "uint8",
            "UInt8",
            "uint16",
            "UInt16",
            "uint32",
            "UInt32",
            "uint64",
            "UInt64",
            "int8",
            "Int8",
            "int16",
            "Int16",
            "int32",
            "Int32",
            "int64",
            "Int64",
            "float32",
            "float64",
        ]

        if version.parse(pd.__version__) >= version.parse('1.2.0'):
            # These DTypes where added in 1.2.0
            data_types.extend(["Float32", "Float64"])
        # Generate a dataframe with all the types, in the form of:
        # b=type1,
        # c=type2
        # ..
        data = {}
        for letter, dtype in zip(string.ascii_lowercase, data_types):
            data[letter] = base_df.a.astype(dtype)

        df = pd.DataFrame.from_dict(data)
        conn = duckdb.connect()
        out_df = conn.execute('select * from df').df()

        # Verify that the types in the out_df are correct
        # FIXME: we don't support outputting pandas specific types (i.e UInt64)
        for letter, item in zip(string.ascii_lowercase, data_types):
            column_name = letter
            assert str(out_df[column_name].dtype) == item.lower()

    def test_pandas_unsigned(self, duckdb_cursor):
        unsigned_types = ['uint8', 'uint16', 'uint32', 'uint64']
        data = numpy.array([0, 1, 2, 3])
        for u_type in unsigned_types:
            round_trip(data, u_type)

    def test_pandas_bool(self, duckdb_cursor):
        data = numpy.array([True, False, False, True])
        round_trip(data, 'bool')

    def test_pandas_boolean(self, duckdb_cursor):
        data = numpy.array([True, None, pd.NA, numpy.nan, True])
        df_in = pd.DataFrame(
            {
                'object': pd.Series(data, dtype='boolean'),
            }
        )

        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()
        assert df_out['object'][0] == df_in['object'][0]
        assert numpy.isnan(df_out['object'][1])
        assert numpy.isnan(df_out['object'][2])
        assert numpy.isnan(df_out['object'][3])
        assert df_out['object'][4] == df_in['object'][4]

    def test_pandas_float32(self, duckdb_cursor):
        data = numpy.array([0.1, 0.32, 0.78, numpy.nan])
        df_in = pd.DataFrame(
            {
                'object': pd.Series(data, dtype='float32'),
            }
        )

        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()

        assert df_out['object'][0] == df_in['object'][0]
        assert df_out['object'][1] == df_in['object'][1]
        assert df_out['object'][2] == df_in['object'][2]
        assert numpy.isnan(df_out['object'][3])

    def test_pandas_float64(self):
        data = numpy.array([0.233, numpy.nan, 3456.2341231, float('-inf'), -23424.45345, float('+inf'), 0.0000000001])
        df_in = pd.DataFrame(
            {
                'object': pd.Series(data, dtype='float64'),
            }
        )
        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()

        for i in range(len(data)):
            if numpy.isnan(df_out['object'][i]):
                assert i == 1
                continue
            assert df_out['object'][i] == df_in['object'][i]

    def test_pandas_interval(self, duckdb_cursor):
        if pd.__version__ != '1.2.4':
            return

        data = numpy.array([2069211000000000, numpy.datetime64("NaT")])
        df_in = pd.DataFrame(
            {
                'object': pd.Series(data, dtype='timedelta64[ns]'),
            }
        )

        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()

        assert df_out['object'][0] == df_in['object'][0]
        assert pd.isnull(df_out['object'][1])

    def test_pandas_encoded_utf8(self, duckdb_cursor):
        data = u'\u00c3'  # Unicode data
        data = [data.encode('utf8')]
        expected_result = data[0]
        df_in = pd.DataFrame({'object': pd.Series(data, dtype='object')})
        result = duckdb.query_df(df_in, "data", "SELECT * FROM data").fetchone()[0]
        assert result == expected_result



import pytest
import duckdb


class TestStreamingResult(object):
    def test_fetch_one(self, duckdb_cursor):
        # fetch one
        res = duckdb_cursor.sql('SELECT * FROM range(100000)')
        result = []
        while len(result) < 5000:
            tpl = res.fetchone()
            result.append(tpl[0])
        assert result == list(range(5000))

        # fetch one with error
        res = duckdb_cursor.sql(
            "SELECT CASE WHEN i < 10000 THEN i ELSE concat('hello', i::VARCHAR)::INT END FROM range(100000) t(i)"
        )
        with pytest.raises(duckdb.ConversionException):
            while True:
                tpl = res.fetchone()
                if tpl is None:
                    break

    def test_fetch_many(self, duckdb_cursor):
        # fetch many
        res = duckdb_cursor.sql('SELECT * FROM range(100000)')
        result = []
        while len(result) < 5000:
            tpl = res.fetchmany(10)
            result += [x[0] for x in tpl]
        assert result == list(range(5000))

        # fetch many with error
        res = duckdb_cursor.sql(
            "SELECT CASE WHEN i < 10000 THEN i ELSE concat('hello', i::VARCHAR)::INT END FROM range(100000) t(i)"
        )
        with pytest.raises(duckdb.ConversionException):
            while True:
                tpl = res.fetchmany(10)
                if tpl is None:
                    break

    def test_record_batch_reader(self, duckdb_cursor):
        pytest.importorskip("pyarrow")
        pytest.importorskip("pyarrow.dataset")
        # record batch reader
        res = duckdb_cursor.sql('SELECT * FROM range(100000) t(i)')
        reader = res.fetch_arrow_reader(batch_size=16_384)
        result = []
        for batch in reader:
            result += batch.to_pydict()['i']
        assert result == list(range(100000))

        # record batch reader with error
        res = duckdb_cursor.sql(
            "SELECT CASE WHEN i < 10000 THEN i ELSE concat('hello', i::VARCHAR)::INT END FROM range(100000) t(i)"
        )
        reader = res.fetch_arrow_reader(batch_size=16_384)
        with pytest.raises(OSError):
            result = []
            for batch in reader:
                result += batch.to_pydict()['i']

    def test_9801(self, duckdb_cursor):
        duckdb_cursor.execute('CREATE TABLE test(id INTEGER , name VARCHAR NOT NULL);')

        words = ['aaaaaaaaaaaaaaaaaaaaaaa', 'bbbb', 'ccccccccc', 'ííííííííí']
        lines = [(i, words[i % 4]) for i in range(1000)]
        duckdb_cursor.executemany("INSERT INTO TEST (id, name) VALUES (?, ?)", lines)

        rel1 = duckdb_cursor.sql(
            """
            SELECT id, name FROM test ORDER BY id ASC
        """
        )
        result = rel1.fetchmany(size=5)
        counter = 0
        while result != []:
            for x in result:
                assert x == (counter, words[counter % 4])
                counter += 1
            result = rel1.fetchmany(size=5)



import duckdb
import os
import pytest

pd = pytest.importorskip("pandas")
pa = pytest.importorskip("pyarrow")
from typing import Union
import pyarrow.compute as pc
import uuid
import datetime
import numpy as np
import cmath

from duckdb.typing import *


def make_annotated_function(type):
    # Create a function that returns its input
    def test_base(x):
        return x

    import types

    test_function = types.FunctionType(
        test_base.__code__, test_base.__globals__, test_base.__name__, test_base.__defaults__, test_base.__closure__
    )
    # Add annotations for the return type and 'x'
    test_function.__annotations__ = {'return': type, 'x': type}
    return test_function


class TestScalarUDF(object):
    @pytest.mark.parametrize('function_type', ['native', 'arrow'])
    @pytest.mark.parametrize(
        'test_type',
        [
            (TINYINT, -42),
            (SMALLINT, -512),
            (INTEGER, -131072),
            (BIGINT, -17179869184),
            (UTINYINT, 254),
            (USMALLINT, 65535),
            (UINTEGER, 4294967295),
            (UBIGINT, 18446744073709551615),
            (HUGEINT, 18446744073709551616),
            (VARCHAR, 'long_string_test'),
            (UUID, uuid.UUID('ffffffff-ffff-ffff-ffff-ffffffffffff')),
            (FLOAT, 0.12246409803628922),
            (DOUBLE, 123142.12312416293784721232344),
            (DATE, datetime.date(2005, 3, 11)),
            (TIMESTAMP, datetime.datetime(2009, 2, 13, 11, 5, 53)),
            (TIME, datetime.time(14, 1, 12)),
            (BLOB, b'\xF6\x96\xB0\x85'),
            (INTERVAL, datetime.timedelta(days=30969, seconds=999, microseconds=999999)),
            (BOOLEAN, True),
            (
                duckdb.struct_type(['BIGINT[]', 'VARCHAR[]']),
                {'v1': [1, 2, 3], 'v2': ['a', 'non-inlined string', 'duckdb']},
            ),
            (duckdb.list_type('VARCHAR'), ['the', 'duck', 'non-inlined string']),
        ],
    )
    def test_type_coverage(self, test_type, function_type):
        type = test_type[0]
        value = test_type[1]

        test_function = make_annotated_function(type)

        con = duckdb.connect()
        con.create_function('test', test_function, type=function_type)

        # Single value
        res = con.execute(f"select test(?::{str(type)})", [value]).fetchall()
        assert res[0][0] == value

        # NULLs
        res = con.execute(f"select res from (select ?, test(NULL::{str(type)}) as res)", [value]).fetchall()
        assert res[0][0] == None

        # Multiple chunks
        size = duckdb.__standard_vector_size__ * 3
        res = con.execute(f"select test(x) from repeat(?::{str(type)}, {size}) as tbl(x)", [value]).fetchall()
        assert len(res) == size

        # Mixed NULL/NON-NULL
        size = duckdb.__standard_vector_size__ * 3
        con.execute("select setseed(0.1337)").fetchall()
        actual = con.execute(
            f"""
            select test(
                case when (x > 0.5) then
                    ?::{str(type)}
                else
                    NULL
                end
            ) from (select random() as x from range({size}))
        """,
            [value],
        ).fetchall()

        con.execute("select setseed(0.1337)").fetchall()
        expected = con.execute(
            f"""
            select
                case when (x > 0.5) then
                    ?::{str(type)}
                else
                    NULL
                end
            from (select random() as x from range({size}))
        """,
            [value],
        ).fetchall()
        assert expected == actual

        # Using 'relation.project'
        con.execute(f"create table tbl as select ?::{str(type)} as x", [value])
        table_rel = con.table('tbl')
        res = table_rel.project('test(x)').fetchall()
        assert res[0][0] == value

    @pytest.mark.parametrize('udf_type', ['arrow', 'native'])
    def test_map_coverage(self, udf_type):
        def no_op(x):
            return x

        con = duckdb.connect()
        map_type = con.map_type('VARCHAR', 'BIGINT')
        con.create_function('test_map', no_op, [map_type], map_type, type=udf_type)
        rel = con.sql("select test_map(map(['non-inlined string', 'test', 'duckdb'], [42, 1337, 123]))")
        res = rel.fetchall()
        assert res == [({'key': ['non-inlined string', 'test', 'duckdb'], 'value': [42, 1337, 123]},)]

    @pytest.mark.parametrize('udf_type', ['arrow', 'native'])
    def test_exceptions(self, udf_type):
        def raises_exception(x):
            raise AttributeError("error")

        con = duckdb.connect()
        con.create_function('raises', raises_exception, [BIGINT], BIGINT, type=udf_type)
        with pytest.raises(
            duckdb.InvalidInputException,
            match=' Python exception occurred while executing the UDF: AttributeError: error',
        ):
            res = con.sql('select raises(3)').fetchall()

        con.remove_function('raises')
        con.create_function(
            'raises', raises_exception, [BIGINT], BIGINT, exception_handling='return_null', type=udf_type
        )
        res = con.sql('select raises(3) from range(5)').fetchall()
        assert res == [(None,), (None,), (None,), (None,), (None,)]

    def test_non_callable(self):
        con = duckdb.connect()
        with pytest.raises(TypeError):
            con.create_function('func', 5, [BIGINT], BIGINT, type='arrow')

        class MyCallable:
            def __init__(self):
                pass

            def __call__(self, x):
                return x

        my_callable = MyCallable()
        con.create_function('func', my_callable, [BIGINT], BIGINT, type='arrow')
        res = con.sql('select func(5)').fetchall()
        assert res == [(5,)]

    # pyarrow does not support creating an array filled with pd.NA values
    @pytest.mark.parametrize('udf_type', ['native'])
    @pytest.mark.parametrize('duckdb_type', [FLOAT, DOUBLE])
    def test_pd_nan(self, duckdb_type, udf_type):
        def return_pd_nan():
            if udf_type == 'native':
                return pd.NA

        con = duckdb.connect()
        con.create_function('return_pd_nan', return_pd_nan, None, duckdb_type, null_handling='SPECIAL', type=udf_type)

        res = con.sql('select return_pd_nan()').fetchall()
        assert res[0][0] == None

    def test_side_effects(self):
        def count() -> int:
            old = count.counter
            count.counter += 1
            return old

        count.counter = 0

        con = duckdb.connect()
        con.create_function('my_counter', count, side_effects=False)
        res = con.sql('select my_counter() from range(10)').fetchall()
        assert res == [(0,), (0,), (0,), (0,), (0,), (0,), (0,), (0,), (0,), (0,)]

        count.counter = 0
        con.remove_function('my_counter')
        con.create_function('my_counter', count, side_effects=True)
        res = con.sql('select my_counter() from range(10)').fetchall()
        assert res == [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,)]

    @pytest.mark.parametrize('udf_type', ['arrow', 'native'])
    @pytest.mark.parametrize('duckdb_type', [FLOAT, DOUBLE])
    def test_np_nan(self, duckdb_type, udf_type):
        def return_np_nan():
            if udf_type == 'native':
                return np.nan
            else:
                import pyarrow as pa

                return pa.chunked_array([[np.nan]], type=pa.float64())

        con = duckdb.connect()
        con.create_function('return_np_nan', return_np_nan, None, duckdb_type, null_handling='SPECIAL', type=udf_type)

        res = con.sql('select return_np_nan()').fetchall()
        assert pd.isnull(res[0][0])

    @pytest.mark.parametrize('udf_type', ['arrow', 'native'])
    @pytest.mark.parametrize('duckdb_type', [FLOAT, DOUBLE])
    def test_math_nan(self, duckdb_type, udf_type):
        def return_math_nan():
            import cmath

            if udf_type == 'native':
                return cmath.nan
            else:
                import pyarrow as pa

                return pa.chunked_array([[cmath.nan]], type=pa.float64())

        con = duckdb.connect()
        con.create_function(
            'return_math_nan', return_math_nan, None, duckdb_type, null_handling='SPECIAL', type=udf_type
        )

        res = con.sql('select return_math_nan()').fetchall()
        assert pd.isnull(res[0][0])

    @pytest.mark.parametrize('udf_type', ['arrow', 'native'])
    @pytest.mark.parametrize(
        'data_type',
        [
            TINYINT,
            SMALLINT,
            INTEGER,
            BIGINT,
            UTINYINT,
            USMALLINT,
            UINTEGER,
            UBIGINT,
            HUGEINT,
            UHUGEINT,
            VARCHAR,
            UUID,
            FLOAT,
            DOUBLE,
            DATE,
            TIMESTAMP,
            TIME,
            BLOB,
            INTERVAL,
            BOOLEAN,
            duckdb.struct_type(['BIGINT[]', 'VARCHAR[]']),
            duckdb.list_type('VARCHAR'),
        ],
    )
    def test_return_null(self, data_type, udf_type):
        def return_null():
            if udf_type == 'native':
                return None
            else:
                import pyarrow as pa

                return pa.nulls(1)

        con = duckdb.connect()
        con.create_function('return_null', return_null, None, data_type, null_handling='special', type=udf_type)
        rel = con.sql('select return_null() as x')
        assert rel.types[0] == data_type
        assert rel.fetchall()[0][0] == None

    def test_udf_transaction_interaction(self):
        def func(x: int) -> int:
            return x

        con = duckdb.connect()
        rel = con.sql('select 42')
        # Using fetchone keeps the result open, with a transaction
        rel.fetchone()

        # If we would allow a UDF to be created when a transaction is active
        # then starting a new result-fetch would cancel the transaction
        # which would corrupt our internal mechanism used to check if a UDF is already registered
        # because that isn't transaction-aware
        with pytest.raises(
            duckdb.InvalidInputException,
            match='This function can not be called with an active transaction!, commit or abort the existing one first',
        ):
            con.create_function('func', func)

        # This would cancel the previous transaction, causing the function to no longer exist
        rel.fetchall()

        con.create_function('func', func)
        res = con.sql('select func(5)').fetchall()
        assert res == [(5,)]



import pytest
import duckdb


class Test3324(object):
    def test_3324(self, duckdb_cursor):
        create_output = duckdb_cursor.execute(
            """
        create or replace table my_table as 
        select 'test1' as column1, 1 as column2, 'quack' as column3 
        union all
        select 'test2' as column1, 2 as column2, 'quacks' as column3 
        union all
        select 'test3' as column1, 3 as column2, 'quacking' as column3 
        """
        ).fetch_df()
        prepare_output = duckdb_cursor.execute(
            """
            prepare v1 as 
                select 
                    column1
                    , column2
                    , column3 
                from my_table
                where 
                    column1 = $1"""
        ).fetch_df()

        with pytest.raises(duckdb.BinderException, match="Unexpected prepared parameter"):
            duckdb_cursor.execute("""execute v1(?)""", 'test1').fetch_df()
        with pytest.raises(duckdb.BinderException, match="Unexpected prepared parameter"):
            duckdb_cursor.execute("""execute v1(?)""", ('test1',)).fetch_df()
